/* ptpcam.c
 *
 * Copyright (C) 2001-2005 Mariusz Woloszyn <emsi@ipartners.pl>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#define PTPCAM_VERSION  "2.2"  /* [sawas-6] */

#undef  DEBUG					/* for debug  [sawas-4] */
#undef  DEBUG_TIME				/* for time measurement  [sawas-4] */
#undef  DEBUG_POLL 		 		/* for polling debug  [sawas-5] */
#undef  DEBUG_SEARCH			/* for search debug  [sawas-55] */
#undef  DEBUG_PARTIAL			/* for partial file debug  [sawas-55] */

#include <config.h>
#include "ptp.h"
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <getopt.h>
#include <unistd.h>
#include <sys/types.h>
#include <utime.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <usb.h>
#ifdef  DEBUG_TIME  /* for time measurement  [sawas-4] */
#include <sys/time.h>		/* gettimeofday()  [sawas-5] */
#endif

#ifdef ENABLE_NLS
#  include <libintl.h>
#  undef _
#  define _(String) dgettext (GETTEXT_PACKAGE, String)
#  ifdef gettext_noop
#    define N_(String) gettext_noop (String)
#  else
#    define N_(String) (String)
#  endif
#else
#  define textdomain(String) (String)
#  define gettext(String) (String)
#  define dgettext(Domain,Message) (Message)
#  define dcgettext(Domain,Message,Type) (Message)
#  define bindtextdomain(Domain,Directory) (Domain)
#  define _(String) (String)
#  define N_(String) (String)
#endif

#include "ptpcam.h"

/* some defines comes here */

/* USB interface class */
#ifndef USB_CLASS_PTP
#define USB_CLASS_PTP		6
#endif

/* USB control message data phase direction */
#ifndef USB_DP_HTD
#define USB_DP_HTD		(0x00 << 7)	/* host to device */
#endif
#ifndef USB_DP_DTH
#define USB_DP_DTH		(0x01 << 7)	/* device to host */
#endif

/* PTP class specific requests */
#ifndef USB_REQ_DEVICE_RESET
#define USB_REQ_DEVICE_RESET		0x66
#endif
#ifndef USB_REQ_GET_DEVICE_STATUS
#define USB_REQ_GET_DEVICE_STATUS	0x67
#endif

/* USB Feature selector HALT */
#ifndef USB_FEATURE_HALT
#define USB_FEATURE_HALT	0x00
#endif

/* OUR APPLICATION USB URB (2MB) ;) */
#define PTPCAM_USB_URB		2097152

#if  0  /* 0: timeout on Canon HV10 GetObject  [sawas-4][SS-3] */
#define USB_TIMEOUT		4000
#else  /* [sawas-4] */
#define USB_TIMEOUT		12000  /* [ms] actual:9s */
#endif  /* [sawas-4] */

#define USB_CAPTURE_TIMEOUT	20000
#define DEFAULT_PARTIAL_SIZE	0x100000  /* default [Byte] [sawas-55] */

#undef GET_PARTIAL_FILE_FOR_OL  /* [sawas-2.112] */
#ifdef GET_PARTIAL_FILE_FOR_OL
#define  VENDOR_ID_OL  0x07B4
#endif

/* one global variable (yes, I know it sucks) */
short verbose=0;
/* the other one, it sucks definitely ;) */
int ptpcam_usb_timeout = USB_TIMEOUT;

#if  1  /* difference from original  [sawas-55] */
/* number of scan port */
static int scan_port  = 2;  /* default */
static int hub_id     = 0;
static int hub_port   = 0;
static int ohci2_port = 0;		/* OHCI  [sawas-6] */
static int use_ohci2_port = 0;  /* OHCI  [sawas-6] */
static unsigned int part_size = DEFAULT_PARTIAL_SIZE;
#endif

void
usage()
{
	printf("USAGE: ptpcam [OPTION]\n\n");
}

void
help()
{
	printf("USAGE: ptpcam [OPTION]\n\n");
	printf("Options:\n"
	"  --bus=BUS-NUMBER             USB bus number\n"
	"  --dev=DEV-NUMBER             USB assigned device number\n"

#if  1  /* difference from original  [sawas-55] */
	"  --hub=HUB-NUMBER             USB assigned hub number\n"
	"  --scan=SCAN PORT-NUMBER      USB scan port number\n"
	"  --port=PORT-NUMBER           USB assigned port number\n"
	"  --part=PARTIAL-SIZE          Partial file size\n"
#endif
	"  -r, --reset                  Reset the device\n"
	"  -l, --list-devices           List all PTP devices\n"
	"  -i, --info                   Show device info\n"
	"  -o, --list-operations        List supported operations\n"
	"  -p, --list-properties        List all PTP device properties\n"
	"                               "
				"(e.g. focus mode, focus distance, etc.)\n"
	"  -s, --show-property=NUMBER   Display property details "
					"(or set its value,\n"
	"                               if used in conjunction with --val)\n"
	"  --set-property=NUMBER        Set property value (--val required)\n"
	"  --val=VALUE                  Property value\n"
	"  -L, --list-files             List all files\n"
	"  -g, --get-file=HANDLE        Get file by given handler\n"
	"  -G, --get-all-files          Get all files\n"
	"  --overwrite                  Force file overwrite while saving"
					"to disk\n"
	"  -d, --delete-object=HANDLE   Delete object (file) by given handle\n"
	"  -D, --delete-all-files       Delete all files form camera\n"
	"  -c, --capture                Initiate capture\n"
	"  --loop-capture=N             Perform N times capture/get/delete\n"

#if  1  /* difference from original  [sawas-4] */
	"  -n, --num-objects=TYPE       Get total num of object by given object type\n"
	"  -I, --get-object-info=HANDLE Get object info by given handler\n"
	"  -E, --list-exif-files        List all EXIF files\n"
	"  -O, --file-operation=DIR     Operation by CommandFile\n"
	"  -t, --get-thum-file=HANDLE   Get Thumbnail file by given handler\n"
#endif

	"  -f, --force                  Talk to non PTP devices\n"
	"  -v, --verbose                Be verbose (print more debug)\n"
	"  -h, --help                   Print this help message\n"
	"\n");
}

#if  1  // [sawas-5]
/*-----------------------------
  Get Port Number  [sawas-5]
   - read /proc/bus/usb/devices
  -----------------------------*/
#define	DESC_NONE		0
#define	DESC_TOPOLOGY	1
#define DESC_PRODUCT	2  /* [sawas-55] */
#define DESC_INTERFACE	3  /* [sawas-55] */

typedef struct ST_DESC_TABLE {
	const char  *parseStr;
	int			descType;
} DESC_TABLE;

DESC_TABLE descTable[] = {
	{"T:", DESC_TOPOLOGY} ,
	{"P:", DESC_PRODUCT},	 /* [sawas-55] */
	{"I:", DESC_INTERFACE},  /* [sawas-55] */
};

int GetDeviceDescType(char *lineBuf)
{
	int i;
	int descType = DESC_NONE;

	for (i = 0; i < sizeof(descTable)/sizeof(DESC_TABLE); i++) {
		const char *parseStr = descTable[i].parseStr;

		if (strlen(lineBuf) > strlen(parseStr) &&
			strncmp(lineBuf, parseStr, strlen(parseStr)) == 0) {
			descType = descTable[i].descType;
		}
	}

	return descType;
}

int SearchDevice(int portno, int busno, int devno)
{
	const char *devfsProc = "/proc/bus/usb/devices";

	int  err = -1;
	FILE *fp = NULL;
	char  lineBuf[256];
	int  descType;
	int  bus = 0, lev, prnt, port, cnt, dev, spd, maxch, ret;
	const char *format_topo
		= "T:  Bus=%d Lev=%d Prnt=%d Port=%d Cnt=%d Dev#=%d Spd=%d MxCh=%d";
#if  1  /* [sawas-55] */
	int  hub_vendor  = 0;
	int  hub_product = 0;
	int  vendId = 0, prodId = 0, revMain, revSub;
	int  hub_dev = 0, dev_prnt = 0;
	const char *format_prod =
		"P:  Vendor=%x ProdID=%x Rev= %x.%x";

	int i, bCont, class;
	char *token, *p_token, *line;
	const char seps[]	= " \t\n\r=()";
	const char *parseStr[] = {
		/*  0 */ "If#",
		/*  1 */ "Alt",
		/*  2 */ "#EPs",
		/*  3 */ "Cls",
		/*  4 */ "Sub",
		/*  5 */ "Prot",
		/*  6 */ "Driver",
	};
#endif

#ifdef  DEBUG_SEARCH
	printf("[ptpcam] %s() IN:\n", __FUNCTION__);
#endif
	fp = fopen(devfsProc, "r");
	if (!fp) {
		printf("Open error %s \n", devfsProc);
		err = -1;
		goto END;
	}

	while (fgets(lineBuf, sizeof(lineBuf), fp)) {  /* [sawas-55] */
		descType = GetDeviceDescType(lineBuf);
		switch (descType) {
		case DESC_TOPOLOGY:
#ifdef  DEBUG_SEARCH
			printf("lineBuf = %s", lineBuf);
#endif
			ret = sscanf(lineBuf, format_topo,
						 &bus, &lev, &prnt, &port, &cnt, &dev, &spd, &maxch);
#ifdef  DEBUG_SEARCH
			printf("Topo(%d): Bus=%d Lev=%d Prnt=%d Port=%d Cnt=%d Dev#=%d Spd=%d MxCh=%d\n",
					ret, bus, lev, prnt, port, cnt, dev, spd, maxch);

			printf("Check: Busno=%d  Portno=%d  Devno=%d\n", busno, portno, devno);  /* [sawas-6] */
#endif

#if  1  /* OHCI  [sawas-6] */
			if (prnt == 0) {  /* Hub */
				continue;
			}

			if (portno == 1) {
				if (ohci2_port) {
					if ((bus  == 3) &&  /* Bus: OHCI2 port */
						(port == 0)) {
						err = 0;  /* found */
						goto CLOSE_END;
					}
				}
				else {
					continue;
				}
			}
			else {  // (portno == 0)
				if (ohci2_port) {
					err = 1;  /* not found */
					goto CLOSE_END;
				}
				else {
					if (bus == 3) {  /* Bus: OHCI2 port */
						continue;
					}
				}
			}
#endif

			if (!hub_port &&
				(port == portno) &&
				(bus  == busno)  &&
				(lev  == 1) &&
				(dev  == devno)) {
#ifdef  DEBUG_SEARCH
				printf("********* PTP device found! (not hub) ******\n");
#endif
				err = 0;  /* found */
				goto CLOSE_END;
			}
			break;

		case DESC_PRODUCT:
			if (!hub_id || !hub_port) {  // not hub port
				continue;
			}

#ifdef  DEBUG_SEARCH
			printf("lineBuf = %s", lineBuf);
#endif
			// use hub?
			if (hub_id && hub_port) { // with hub, hub port
				// get proc ID
				ret = sscanf(lineBuf, format_prod, &vendId, &prodId, &revMain, &revSub);
#ifdef  DEBUG_SEARCH
				printf("vendId:0x%x prodId:0x%x rev=%x.%x\n", vendId, prodId, revMain, revSub);
#endif
			}
			break;

		case DESC_INTERFACE:
			if (!(hub_id && hub_port)) {  // not hub port
				continue;
			}
			
#ifdef  DEBUG_SEARCH
			printf("lineBuf = %s", lineBuf);
#endif
			bCont = 1;
			line = (char *)lineBuf;
			while (bCont && (token = strtok_r(line, seps, &p_token))) {
				//printf("[Token:%d] %s\n", j++, token);
				// for next loop.
				line = NULL;
				for (i = 0; i < sizeof(parseStr)/sizeof(char*); i++) {
					if (strlen(token) >= strlen(parseStr[i]) && 
						strncmp(token, parseStr[i], strlen(parseStr[i])) == 0) {
						// get next token.
						token = strtok_r(NULL, seps, &p_token);
						if (!token) {
							bCont = 0;
							break;
						}

						switch (i) {
						case 3:  // Cls
							class = strtoul(token, NULL, 16);
#ifdef  DEBUG_SEARCH
							printf("class = 0x%x\n", class);
#endif
							token = strtok_r(NULL, seps, &p_token);
							break;

						default:
							break;
						}
					}
				}
			}
			break;

		default:
			err = 1;
			break;
		}

		// class check
		switch (class) {
		case 0x06:  /* Still Image Capture Device */
			if (!((port == portno) &&
				  (bus  == busno)  &&
				  (lev  == 2) &&   // hub-port: level=2
				  (dev  == devno))) {
				continue;
			}
			dev_prnt = prnt;

#ifdef  DEBUG_SEARCH
			printf("********* PTP device: prnt=%d ******\n", prnt);
#endif
			break;

		case 0x09:  /* USB HUB */
#ifdef  DEBUG_SEARCH
			printf("class = 0x%x hub_port:0x%02x\n", class, hub_port);  /* [sawas-6] */
#endif
			// hub ID
			hub_vendor  = (hub_id>>16);			/* Vendor:   upper 16bits  [sawas-6] */
			hub_product = (hub_id & 0xFFFF);	/* Products: lower 16bits  [sawas-6] */
#ifdef  DEBUG_SEARCH
			printf("vendor:0x%x product:0x%x\n", hub_vendor, hub_product);
#endif

			// hub ID check
			if (!((hub_vendor  == vendId) &&
				  (hub_product == prodId) &&
				  (lev == 1))) {   // internal hub: level=1
				continue;
			}
			hub_dev = dev;

#ifdef  DEBUG_SEARCH
			printf("********* HUB device: dev=%d ******\n", dev);
#endif
			break;

		default:
			continue;
		}

#ifdef  DEBUG_SEARCH
		printf("********* dev_prnt=%d hub_dev=%d ******\n", dev_prnt, hub_dev);
#endif
		if (!dev_prnt ||
			!hub_dev  ||
			(dev_prnt != hub_dev)) {
			continue;
		}

#ifdef  DEBUG_SEARCH
		printf("********* PTP device found (under hub) ******\n");
#endif
		err = 0;
		goto CLOSE_END;
	}

 CLOSE_END:
	if (fp) {
		fclose(fp);
	}

 END:
#ifdef  DEBUG_SEARCH
	printf("[ptpcam] %s() END: hub_port=0x%02x, ohci2_port=0x%02x, ret=%d\n",
						__FUNCTION__, hub_port, use_ohci2_port, err);  /* OHCI  [sawas-6] */
#endif
	return err;
}
#endif

static short
ptp_read_func (unsigned char *bytes, unsigned int size, void *data)
{
	int result=-1;
	PTP_USB *ptp_usb=(PTP_USB *)data;
	int toread=0;
	signed long int rbytes=size;

	do {
		bytes+=toread;
		if (rbytes>PTPCAM_USB_URB) 
			toread = PTPCAM_USB_URB;
		else
			toread = rbytes;
		result=USB_BULK_READ(ptp_usb->handle, ptp_usb->inep,(char *)bytes, toread,ptpcam_usb_timeout);
		/* sometimes retry might help */
		if (result==0)
			result=USB_BULK_READ(ptp_usb->handle, ptp_usb->inep,(char *)bytes, toread,ptpcam_usb_timeout);
		if (result < 0)
			break;
		rbytes-=PTPCAM_USB_URB;
	} while (rbytes>0);

	if (result >= 0) {
		return (PTP_RC_OK);
	}
	else 
	{
		if (verbose) perror("usb_bulk_read");
		return PTP_ERROR_IO;
	}
}

static short
ptp_write_func (unsigned char *bytes, unsigned int size, void *data)
{
	int result;
	PTP_USB *ptp_usb=(PTP_USB *)data;

	result=USB_BULK_WRITE(ptp_usb->handle,ptp_usb->outep,(char *)bytes,size,ptpcam_usb_timeout);
	if (result >= 0)
		return (PTP_RC_OK);
	else 
	{
		if (verbose) perror("usb_bulk_write");
		return PTP_ERROR_IO;
	}
}

static short
ptp_check_int (unsigned char *bytes, unsigned int size, void *data)
{
	int result;
	PTP_USB *ptp_usb=(PTP_USB *)data;

	if (verbose) printf ("Awaiting event...\n");

	result=USB_BULK_READ(ptp_usb->handle, ptp_usb->intep,(char *)bytes,size,ptpcam_usb_timeout);
	if (result==0)
		result = USB_BULK_READ(ptp_usb->handle, ptp_usb->intep,(char *) bytes, size, ptpcam_usb_timeout);
	if (result >= 0) {
		return (PTP_RC_OK);
	} else {
		if (verbose) perror("ptp_check_int");
		return PTP_ERROR_IO;
	}
}


void
debug (void *data, const char *format, va_list args);
void
debug (void *data, const char *format, va_list args)
{
	if (verbose<2) return;
	vfprintf (stderr, format, args);
	fprintf (stderr,"\n");
	fflush(stderr);
}

void
error (void *data, const char *format, va_list args);
void
error (void *data, const char *format, va_list args)
{
/*	if (!verbose) return; */
	vfprintf (stderr, format, args);
	fprintf (stderr,"\n");
	fflush(stderr);
}



void
init_ptp_usb (PTPParams* params, PTP_USB* ptp_usb, struct usb_device* dev)
{
	usb_dev_handle *device_handle;

	params->write_func=ptp_write_func;
	params->read_func=ptp_read_func;
	params->check_int_func=ptp_check_int;
	params->check_int_fast_func=ptp_check_int;
	params->error_func=error;
	params->debug_func=debug;
	params->sendreq_func=ptp_usb_sendreq;
	params->senddata_func=ptp_usb_senddata;
	params->getresp_func=ptp_usb_getresp;
	params->getdata_func=ptp_usb_getdata;
	params->data=ptp_usb;
	params->transaction_id=0;
	params->byteorder = PTP_DL_LE;

	if ((device_handle=usb_open(dev))){
		if (!device_handle) {
			perror("usb_open()");
			exit(0);
		}
		ptp_usb->handle=device_handle;
		usb_claim_interface(device_handle,
			dev->config->interface->altsetting->bInterfaceNumber);
	}
}

void
clear_stall(PTP_USB* ptp_usb, struct usb_device* dev)
{
	uint16_t status=0;
	int ret;

	/* check the inep status */
	ret=usb_get_endpoint_status(ptp_usb,ptp_usb->inep,&status);

#if  0  /* difference from original  [sawas-4] */
	if (ret<0) perror ("inep: usb_get_endpoint_status()");
#else
	if (ret<0) {
		printf("clear_stall( ret:[0]%08lx, [1]%08lx, [2]%08lx, [3]%08lx )\n",
			   (long)__builtin_return_address(0),
			   (long)__builtin_return_address(1),
			   (long)__builtin_return_address(2),
			   (long)__builtin_return_address(3));
		perror ("inep: usb_get_endpoint_status()");
	}
#endif
	/* and clear the HALT condition if happend */
	else if (status) {
		printf("Resetting input pipe!\n");
		ret=usb_clear_stall_feature(ptp_usb,ptp_usb->inep);
			/*usb_clear_halt(ptp_usb->handle,ptp_usb->inep); */
		if (ret<0)perror ("usb_clear_stall_feature()");
	}
	status=0;

	/* check the outep status */
	ret=usb_get_endpoint_status(ptp_usb,ptp_usb->outep,&status);
	if (ret<0) perror ("outep: usb_get_endpoint_status()");
	/* and clear the HALT condition if happend */
	else if (status) {
		printf("Resetting output pipe!\n");
			ret=usb_clear_stall_feature(ptp_usb,ptp_usb->outep);
		/*usb_clear_halt(ptp_usb->handle,ptp_usb->outep); */
		if (ret<0)perror ("usb_clear_stall_feature()");
	}

		/*usb_clear_halt(ptp_usb->handle,ptp_usb->intep); */
}

void
close_usb(PTP_USB* ptp_usb, struct usb_device* dev)
{
#if  0  /* ptpcam-memory-leak-fix  [SS-7] */
	clear_stall(ptp_usb, dev);
		usb_release_interface(ptp_usb->handle,
				dev->config->interface->altsetting->bInterfaceNumber);
		usb_close(ptp_usb->handle);
#else
	if(ptp_usb && dev) {
		clear_stall(ptp_usb, dev);
		usb_release_interface(ptp_usb->handle,
			dev->config->interface->altsetting->bInterfaceNumber);
	}
	if(ptp_usb){
		usb_close(ptp_usb->handle);
	}
#endif
}


struct usb_bus*
init_usb()
{
	usb_init();
	usb_find_busses();
	usb_find_devices();
	return (usb_get_busses());
}

/*
   find_device() returns the pointer to a usb_device structure matching
   given busn, devicen numbers. If any or both of arguments are 0 then the
   first matching PTP device structure is returned. 
*/
struct usb_device*
find_device (int portn, int busn, int devicen, short force);  /* [sawas-5] */
struct usb_device*
find_device (int portn, int busn, int devn, short force)  /* [sawas-5] */
{
	struct usb_bus *bus;
	struct usb_device *dev;

	/* printf("[ptpcam] %s(): {port: 0x%02x bus:%d dev:%d force:%d}\n", __FUNCTION__, portn, busn, devn, force);  [sawas-6] */
	bus=init_usb();
	for (; bus; bus = bus->next) {
		for (dev = bus->devices; dev; dev = dev->next) {
			if ((dev->config->interface->altsetting->bInterfaceClass==
				 USB_CLASS_PTP)||force){
				if (dev->descriptor.bDeviceClass!=USB_CLASS_HUB) {
					int curbusn, curdevn;

					curbusn=strtol(bus->dirname,NULL,10);
					curdevn=strtol(dev->filename,NULL,10);

#ifdef  DEBUG_POLL  // [sawas-5]
					printf("cbus=%d port=%d cdev=%d hubport=%d ohci2_port=0x%02x\n",
							curbusn, portn, curdevn, hub_port, ohci2_port);  /* OHCI  [sawas-6] */
#endif
					if (devn==0) {
						if (busn==0) {
							/* check port number */
							if (SearchDevice(portn, curbusn, curdevn) != 0) {
#ifdef  DEBUG_POLL  // [sawas-6]
								printf("find_device(): skip.\n");
								continue;
							}
							printf("find_device(): found.\n");
#else
								continue;
							}
#endif
							return dev;
						}
						if (curbusn==busn) return dev;
					} else {
						if ((busn==0)&&(curdevn==devn)) return dev;
						if ((curbusn==busn)&&(curdevn==devn)) return dev;
					}
				}
			} // if (PTP)
		} // end for(dev)
	} // end for(bus)

	return NULL;
}

void
find_endpoints(struct usb_device *dev, int* inep, int* outep, int* intep);
void
find_endpoints(struct usb_device *dev, int* inep, int* outep, int* intep)
{
	int i,n;
	struct usb_endpoint_descriptor *ep;

	ep = dev->config->interface->altsetting->endpoint;
	n=dev->config->interface->altsetting->bNumEndpoints;

	for (i=0;i<n;i++) {
	if (ep[i].bmAttributes==USB_ENDPOINT_TYPE_BULK)	{
		if ((ep[i].bEndpointAddress&USB_ENDPOINT_DIR_MASK)==
			USB_ENDPOINT_DIR_MASK)
		{
			*inep=ep[i].bEndpointAddress;
			if (verbose)
				printf ("Found inep: 0x%02x\n",*inep);
		}
		if ((ep[i].bEndpointAddress&USB_ENDPOINT_DIR_MASK)==0)
		{
			*outep=ep[i].bEndpointAddress;
			if (verbose)
				printf ("Found outep: 0x%02x\n",*outep);
		}
		} else if (ep[i].bmAttributes==USB_ENDPOINT_TYPE_INTERRUPT){
			if ((ep[i].bEndpointAddress&USB_ENDPOINT_DIR_MASK)==
				USB_ENDPOINT_DIR_MASK)
			{
				*intep=ep[i].bEndpointAddress;
				if (verbose)
					printf ("Found intep: 0x%02x\n",*intep);
			}
		}
	}
}

int
open_camera (int portn, int busn, int devn, short force, PTP_USB *ptp_usb, PTPParams *params, struct usb_device **dev)  /* [sawas-5] */
{
#ifdef DEBUG
	printf("%s: dev#%i bus#%i\n", __FUNCTION__, devn, busn);  /* [sawas-6] */
#endif
	
	*dev=find_device(portn,busn,devn,force);  /* [sawas-5] */
	if (*dev==NULL) {
#if  0  /* ptpcam-memory-leak-fix  [SS-7] */
		fprintf(stderr,"could not find any device matching given "
		"bus/dev numbers\n");
		exit(-1);
#else
		return -1;
#endif
	}
	find_endpoints(*dev,&ptp_usb->inep,&ptp_usb->outep,&ptp_usb->intep);

	init_ptp_usb(params, ptp_usb, *dev);
	if (ptp_opensession(params,1)!=PTP_RC_OK) {
		fprintf(stderr,"ERROR: Could not open session!\n");
		close_usb(ptp_usb, *dev);
		return -1;
	}
	return 0;
}

void
close_camera (PTP_USB *ptp_usb, PTPParams *params, struct usb_device *dev)
{
#if  0  /* ptpcam-memory-leak-fix  [SS-7] */
	if (ptp_closesession(params)!=PTP_RC_OK)
		fprintf(stderr,"ERROR: Could not close session!\n");
#else
	if(params){
		if (ptp_closesession(params)!=PTP_RC_OK)
			fprintf(stderr,"ERROR: Could not close session!\n");
	}
#endif
	close_usb(ptp_usb, dev);
}


void
list_devices(short force)
{
	struct usb_bus *bus;
	struct usb_device *dev;
	int found=0;


	bus=init_usb();
  	for (; bus; bus = bus->next)
		for (dev = bus->devices; dev; dev = dev->next) {
		/* if it's a PTP device try to talk to it */
		if ((dev->config->interface->altsetting->bInterfaceClass==
			USB_CLASS_PTP)||force)
		if (dev->descriptor.bDeviceClass!=USB_CLASS_HUB)
		{
			int n;
			struct usb_endpoint_descriptor *ep;
			PTPParams params;
			PTP_USB ptp_usb;
			PTPDeviceInfo deviceinfo;

#if  1  /* ptpcam-data-initialization  [SS-6] */
			memset(&params, 0, sizeof(params));
#endif
			if (!found){
				printf("\nListing devices...\n");
				printf("bus/dev\tvendorID/prodID\tdevice model\n");
				found=1;
			}
			ep = dev->config->interface->altsetting->endpoint;
			n=dev->config->interface->altsetting->bNumEndpoints;

			find_endpoints(dev,&ptp_usb.inep,&ptp_usb.outep,
				&ptp_usb.intep);
			init_ptp_usb(&params, &ptp_usb, dev);

			CC(ptp_opensession (&params,1),
				"Could not open session!\n"
				"Try to reset the camera.\n");
			CC(ptp_getdeviceinfo (&params, &deviceinfo),
				"Could not get device info!\n");

	  			printf("%s/%s\t0x%04X/0x%04X\t%s\n",
				bus->dirname, dev->filename,
				dev->descriptor.idVendor,
				dev->descriptor.idProduct, deviceinfo.Model);

#if  0  /* ptpcam-memory-leak-fix  [SS-7] */
			CC(ptp_closesession(&params),
				"Could not close session!\n");
			close_usb(&ptp_usb, dev);
#else
			if(dev){
				ptp_free_deviceinfo(&deviceinfo);
				close_camera(&ptp_usb, &params, dev);
			}
#endif
		}
	}
	if (!found) printf("\nFound no PTP devices\n");
	printf("\n");
}

void
show_info (int portn, int busn, int devn, short force, int resfd)  /* [sawas-5] */
{
	PTPParams params;
	PTP_USB ptp_usb;
	struct usb_device *dev;
#if  1  /* ptpcam-memory-leak-fix  [SS-7] */
	int errcode = -1;
#endif

#ifdef  DEBUG_TIME  /* for time measurement  [sawas-4] */
	struct timeval tv, tv_now, tv_ini;

	printf("[ptp/%s] %s(): bus=%i dev=%i force=%d\n", __FILE__, __FUNCTION__, busn, devn, force);
	/* save initial time */
	gettimeofday(&tv_now, NULL);
	tv_ini.tv_sec = tv_now.tv_sec;
	tv_ini.tv_usec = tv_now.tv_usec;
#else  // [sawas-5]
	//printf("[ptp/%s] %s(): bus=%i dev=%i force=%d\n", __FILE__, __FUNCTION__, busn, devn, force);
#endif  /* [sawas-4] */

	memset(&params, 0, sizeof(params));
	printf("\nCamera information\n");
	printf("==================\n");

#if  0  /* ptpcam-memory-leak-fix  [SS-7] */
	if (open_camera(busn, devn, force, &ptp_usb, &params, &dev)<0)
		return;
	CR(ptp_getdeviceinfo (&params, &params.deviceinfo),
		"Could not get device info\n");
#else

	if (open_camera(portn, busn, devn, force, &ptp_usb, &params, &dev)<0){  /* [sawas-5] */
#if  0  /* difference from original  [sawas-4] */
		printf("show_info : Could not open_camera\n");
#else
		fprintf(stderr,"ERROR(%d): Could not open camera(dev:%p).\n", errcode,dev);
#endif
		dev = NULL;
		goto out;
	}
	errcode = ptp_getdeviceinfo(&params, &params.deviceinfo);
	if(errcode != PTP_RC_OK){
		fprintf(stderr,"ERROR(%x): Could not get device info.\n", errcode);
		goto out;
	}
#endif

	printf("Model: %s\n",params.deviceinfo.Model);
	printf("  manufacturer: %s\n",params.deviceinfo.Manufacturer);
	printf("  serial number: '%s'\n",params.deviceinfo.SerialNumber);
	printf("  device version: %s\n",params.deviceinfo.DeviceVersion);
	printf("  extension ID: 0x%08x\n",
					params.deviceinfo.VendorExtensionID);
	printf("  extension description: %s\n",
					params.deviceinfo.VendorExtensionDesc);
	printf("  extension version: 0x%04x\n",
				params.deviceinfo.VendorExtensionVersion);
	printf("\n");

#if  0  /* ptpcam-memory-leak-fix  [SS-7] */
	close_camera(&ptp_usb, &params, dev);
#else
out:
#if  1  /* difference from original  [sawas-4] */
	if(resfd >= 0){
		const char *WRITE_FMT = "N:%d E:%x %s\n";  /* for Part-OBJ  [sawas-55] */
		char buf[BUF_SIZE];  /* for Part-OBJ  [sawas-55] */
		int ret;

		memset(buf, 0, sizeof(buf));
		if(errcode == PTP_RC_OK){
			snprintf(buf, sizeof(buf)-1, WRITE_FMT,
					 0, errcode, params.deviceinfo.Manufacturer);
			ret = write(resfd, buf, strlen(buf));

			memset(buf, 0, sizeof(buf));
			snprintf(buf, sizeof(buf)-1, WRITE_FMT,
					 1, errcode, params.deviceinfo.Model);
			ret = write(resfd, buf, strlen(buf));

			if(ret != strlen(buf)){
				printf("# ptpcam : write device info error(%d)\n", ret);
			}
		}
		else {
			snprintf(buf, sizeof(buf)-1, WRITE_FMT,
					 0, errcode, "not found any device.");
			write(resfd, buf, strlen(buf));
		}
	}
#endif
	if(dev){
		ptp_free_deviceinfo(&params.deviceinfo);
		close_camera(&ptp_usb, &params, dev);
	}

#ifdef  DEBUG_TIME  /* for time measurement  [sawas-4] */
	printf("[ptp/%s] %s(): done\n", __FILE__, __FUNCTION__);

time:
	/* print time */
	gettimeofday(&tv_now, NULL);
	tv.tv_usec = tv_now.tv_usec - tv_ini.tv_usec;
	tv.tv_sec = tv_now.tv_sec - tv_ini.tv_sec;

	if (tv.tv_usec < 0) {
		tv.tv_usec += 1000000;
		tv.tv_sec--;
	}

	printf("\n[ptp/%s] %s(): ######  Time = %lds %ldms  ######\n",
					__FILE__, __FUNCTION__, tv.tv_sec, tv.tv_usec/1000);
#endif
	return;
#endif
}

void
capture_image (int portn, int busn, int devn, short force)  /* [sawas-5] */
{
	PTPParams params;
	PTP_USB ptp_usb;
	PTPContainer event;
	struct usb_device *dev;
	short ret;

#if  1  /* ptpcam-data-initialization  [SS-6] */
	memset(&params, 0, sizeof(params));
#endif

#if  0  /* ptpcam-memory-leak-fix  [SS-7] */
	printf("\nInitiating captue...\n");
	if (open_camera(busn, devn, force, &ptp_usb, &params, &dev)<0)
		return;
#else
	printf("\nInitiating capture...\n");
	if (open_camera(portn, busn, devn, force, &ptp_usb, &params, &dev)<0){  /* [sawas-5] */
		printf("capture_image : Could not open_camera\n");
		dev = NULL;
		goto out;
	}
#endif

	/* capture timeout should be longer */
	ptpcam_usb_timeout=USB_CAPTURE_TIMEOUT;

	CR(ptp_initiatecapture (&params, 0x0, 0), "Could not capture.\n");
	
	ret=ptp_usb_event_wait(&params,&event);
	if (verbose) printf ("Event received %08x, ret=%x\n", event.Code, ret);
	if (ret!=PTP_RC_OK) goto err;
	if (event.Code==PTP_EC_CaptureComplete) {
		printf ("Camera reported 'capture completed' but the object information is missing.\n");
		goto out;
	}
		
	while (event.Code==PTP_EC_ObjectAdded) {
		printf ("Object added 0x%08x\n", event.Param1);
		if (ptp_usb_event_wait(&params, &event)!=PTP_RC_OK)
			goto err;
		if (verbose) printf ("Event received %08x, ret=%x\n", event.Code, ret);
		if (event.Code==PTP_EC_CaptureComplete) {
			printf ("Capture completed successfully!\n");
			goto out;
		}
	}
	
err:
	printf("Events receiving error. Capture status unknown.\n");
out:

	ptpcam_usb_timeout=USB_TIMEOUT;
#if  0  /* ptpcam-memory-leak-fix  [SS-7] */
	close_camera(&ptp_usb, &params, dev);
#else
	if(dev){
		close_camera(&ptp_usb, &params, dev);
	}
#endif
}

void
loop_capture (int portn, int busn, int devn, short force, int n,  int overwrite)  /* [sawas-5] */
{
	PTPParams params;
	PTP_USB ptp_usb;
	PTPContainer event;
	struct usb_device *dev;
	int file;
	PTPObjectInfo oi;
	uint32_t handle=0;
	char *image;
	int ret;
	char *filename;
#if  1  /* ptpcam-memory-leak-fix  [SS-7] */
	int errcode = -1;
#endif

#if  1  /* ptpcam-data-initialization  [SS-6] */
	memset(&params, 0, sizeof(params));
	memset(&oi, 0, sizeof(oi));
#endif

#if  0  /* ptpcam-memory-leak-fix  [SS-7] */
	if (open_camera(busn, devn, force, &ptp_usb, &params, &dev)<0)
		return;
#else
	if (open_camera(portn, busn, devn, force, &ptp_usb, &params, &dev)<0){  /* [sawas-5] */
		printf("loop_capture : Could not open_camera\n");
		dev = NULL;
		goto err;
	}
#endif

	/* capture timeout should be longer */
	ptpcam_usb_timeout=USB_CAPTURE_TIMEOUT;

#if  0  /* ptpcam-memory-leak-fix  [SS-7] */
	CR(ptp_getdeviceinfo (&params, &params.deviceinfo),
		"Could not get device info\n");
#else
	errcode = ptp_getdeviceinfo(&params, &params.deviceinfo);
	if(errcode != PTP_RC_OK){
		fprintf(stderr,"ERROR(%x): Could not get device info.\n", errcode);
		goto err;
	}
#endif
	printf("Camera: %s\n",params.deviceinfo.Model);


	/* local loop */
	while (n>0) {
		/* capture */
		printf("\nInitiating captue...\n");
		CR(ptp_initiatecapture (&params, 0x0, 0),"Could not capture\n");
		n--;

		ret=ptp_usb_event_wait(&params,&event);
		if (verbose) printf ("Event received %08x, ret=%x\n", event.Code, ret);
		if (ret!=PTP_RC_OK) goto err;
		if (event.Code==PTP_EC_CaptureComplete) {
			printf ("CANNOT DOWNLOAD: got 'capture completed' but the object information is missing.\n");
			goto out;
		}
			
		while (event.Code==PTP_EC_ObjectAdded) {
			printf ("Object added 0x%08x\n", event.Param1);
			handle=event.Param1;
			if (ptp_usb_event_wait(&params, &event)!=PTP_RC_OK)
				goto err;
			if (verbose) printf ("Event received %08x, ret=%x\n", event.Code, ret);
			if (event.Code==PTP_EC_CaptureComplete)
				goto download;
		}
download:	

		memset(&oi, 0, sizeof(PTPObjectInfo));
		if (verbose) printf ("Downloading: 0x%08x\n",handle);
		if ((ret=ptp_getobjectinfo(&params,handle, &oi))!=PTP_RC_OK){
			fprintf(stderr,"ERROR: Could not get object info\n");
			ptp_perror(&params,ret);
			if (ret==PTP_ERROR_IO) clear_stall(&ptp_usb, dev);
			continue;
		}
	
		if (oi.ObjectFormat == PTP_OFC_Association)
				goto out;
		filename=(oi.Filename);
		file=open(filename, (overwrite==OVERWRITE_EXISTING?0:O_EXCL)|O_RDWR|O_CREAT|O_TRUNC,S_IRWXU|S_IRGRP);
		if (file==-1) {
			if (errno==EEXIST) {
				printf("Skipping file: \"%s\", file exists!\n",filename);
				goto out;
			}
			perror("open");
			goto out;
		}
		lseek(file,oi.ObjectCompressedSize-1,SEEK_SET);
		ret=write(file,"",1);
		if (ret==-1) {
			perror("write");
			goto out;
		}
		image=mmap(0,oi.ObjectCompressedSize,PROT_READ|PROT_WRITE,MAP_SHARED,
			file,0);
		if (image==MAP_FAILED) {
			perror("mmap");
			close(file);
			goto out;
		}
		printf ("Saving file: \"%s\" ",filename);
		fflush(NULL);
		ret=ptp_getobject(&params,handle,&image);
		munmap(image,oi.ObjectCompressedSize);
		close(file);
		if (ret!=PTP_RC_OK) {
			printf ("error!\n");
			ptp_perror(&params,ret);
			if (ret==PTP_ERROR_IO) clear_stall(&ptp_usb, dev);
		} else {
			/* and delete from camera! */
			printf("is done...\nDeleting from camera.\n");

#if  0  /* ptpcam-memory-leak-fix  [SS-7] */
			CR(ptp_deleteobject(&params, handle,0),
					"Could not delete object\n");
#else
			errcode = ptp_deleteobject(&params, handle, 0);
			if(errcode != PTP_RC_OK){
				fprintf(stderr,"ERROR(%x): Could not delete object.\n", errcode);
				goto out;
			}
#endif
			printf("Object 0x%08x (%s) deleted.\n",handle, oi.Filename);
		}

#if  0  /* ptpcam-memory-leak-fix  [SS-7] */
out:
		;
	}
err:

	ptpcam_usb_timeout=USB_TIMEOUT;
	close_camera(&ptp_usb, &params, dev);
#else
out:
		ptp_free_objectinfo(&oi);
	}
err:
	ptpcam_usb_timeout=USB_TIMEOUT;
	if(dev){
		ptp_free_deviceinfo(&params.deviceinfo);
		close_camera(&ptp_usb, &params, dev);
	}
#endif
}

void
list_files (int portn, int busn, int devn, short force)  /* [sawas-5] */
{
	PTPParams params;
	PTP_USB ptp_usb;
	struct usb_device *dev;
	int i;
	PTPObjectInfo oi;
#if  1  /* ptpcam-memory-leak-fix  [SS-7] */
	int errcode = -1;
#endif

#if  1  /* ptpcam-data-initialization  [SS-6] */
	memset(&params, 0, sizeof(params));
	memset(&oi, 0, sizeof(oi));
#endif
	printf("\nListing files...\n");

#if  0  /* ptpcam-memory-leak-fix  [SS-7] */
	if (open_camera(busn, devn, force, &ptp_usb, &params, &dev)<0)
		return;
	CR(ptp_getdeviceinfo (&params, &params.deviceinfo),
		"Could not get device info\n");
	printf("Camera: %s\n",params.deviceinfo.Model);
	CR(ptp_getobjecthandles (&params,0xffffffff, 0x000000, 0x000000,
		&params.handles),"Could not get object handles\n");
#else
	if (open_camera(portn, busn, devn, force, &ptp_usb, &params, &dev)<0){  /* [sawas-5] */
		printf("list_files : Could not open_camera\n");
		dev = NULL;
		goto out;
	}
	errcode = ptp_getdeviceinfo(&params, &params.deviceinfo);
	if(errcode != PTP_RC_OK){
		fprintf(stderr,"ERROR(%x): Could not get device info.\n", errcode);
		goto out;
	}
	printf("Camera: %s\n",params.deviceinfo.Model);
	errcode = ptp_getobjecthandles (&params,0xffffffff, 0x000000, 0x000000,
		&params.handles);
	if(errcode != PTP_RC_OK){
		fprintf(stderr,"ERROR(%x): Could not get object handles\n", errcode);
		goto out;
	}
#endif
	printf("Handler:        size: \tname:\n");
	for (i = 0; i < params.handles.n; i++) {
#if  1  /* ptpcam-data-initialization  [SS-6] */
		memset(&oi, 0, sizeof(oi));
#endif

#if  0  /* ptpcam-memory-leak-fix  [SS-7] */
		CR(ptp_getobjectinfo(&params,params.handles.Handler[i],
			&oi),"Could not get object info\n");
		if (oi.ObjectFormat == PTP_OFC_Association)
			continue;
		printf("0x%08x: % 9i\t%s\n",params.handles.Handler[i],
			oi.ObjectCompressedSize, oi.Filename);
	}
	printf("\n");
	close_camera(&ptp_usb, &params, dev);
#else
		errcode = ptp_getobjectinfo(&params,params.handles.Handler[i],&oi);
		if(errcode != PTP_RC_OK){
			fprintf(stderr,"ERROR(%x): Could not get object info.\n", errcode);
			goto out;
		}
		if (oi.ObjectFormat == PTP_OFC_Association){
			ptp_free_objectinfo(&oi);
			continue;
		}
		printf("0x%08x: % 9i\t%s\n",params.handles.Handler[i],
			oi.ObjectCompressedSize, oi.Filename);
		ptp_free_objectinfo(&oi);
	}
	printf("\n");
out:
	if(dev){
		ptp_free_deviceinfo(&params.deviceinfo);
		ptp_free_objecthandles(&params.handles);
		close_camera(&ptp_usb, &params, dev);
	}
#endif
}

#if  1  /* difference from original  [sawas-4] */
void
num_of_objects (int portn, int busn, int devn, short force, uint32_t ofc, int resfd)
{
	PTPParams params;
	PTP_USB ptp_usb;
	struct usb_device *dev = NULL;
	int errcode = -1;
	uint32_t numobjects=0;
	PTPObjectInfo tmpoi;

#ifdef  DEBUG_TIME  /* for time measurement  [sawas-4] */
	struct timeval tv, tv_now, tv_ini;

	printf("[ptp/%s] %s(): port=%i bus=%i dev=%i force=%d\n", __FILE__, __FUNCTION__, portn, busn, devn, force);
	/* save initial time */
	gettimeofday(&tv_now, NULL);
	tv_ini.tv_sec = tv_now.tv_sec;
	tv_ini.tv_usec = tv_now.tv_usec;
#endif  /* [sawas-4] */

	memset(&params, 0, sizeof(params));
	memset(&tmpoi, 0, sizeof(tmpoi));

	if (open_camera(portn, busn, devn, force, &ptp_usb, &params, &dev)<0){  /* [sawas-5] */
		printf("num_of_object : Could not open_camera\n");
		close_camera(NULL,NULL,NULL);
		dev = NULL;
		goto out;
	}

	errcode = ptp_getdeviceinfo(&params, &params.deviceinfo);
	if(errcode != PTP_RC_OK){
		fprintf(stderr,"ERROR(%x): Could not get device info.\n", errcode);
		goto out;
	}

#if 0
	// not work...
	errcode = ptp_getnumobjects (&params,0xffffffff, ofc, 0x000000,
								 &numobjects);
#else
	errcode = ptp_getobjecthandles (&params,0xffffffff, ofc, 0x000000,
									&params.handles);
#endif
	if(errcode != PTP_RC_OK){
		fprintf(stderr,"ERROR(%x): get num of objects.\n", errcode);
		goto out;
	}
	numobjects = params.handles.n;
	// printf("\nNum Of Objects %d (OFC=0x%08x)\n", numobjects, ofc);
 out:
	if(resfd >= 0){
		write_result(resfd, numobjects, errcode, 0, &tmpoi);
	}
	if(dev){
		ptp_free_deviceinfo(&params.deviceinfo);
		ptp_free_objecthandles(&params.handles);
		close_camera(&ptp_usb, &params, dev);
	}

#ifdef  DEBUG_TIME  /* for time measurement  [sawas-4] */
	printf("[ptp/%s] %s(): done\n", __FILE__, __FUNCTION__);

	/* print time */
	gettimeofday(&tv_now, NULL);
	tv.tv_usec = tv_now.tv_usec - tv_ini.tv_usec;
	tv.tv_sec = tv_now.tv_sec - tv_ini.tv_sec;

	if (tv.tv_usec < 0) {
		tv.tv_usec += 1000000;
		tv.tv_sec--;
	}

	printf("\n[ptp/%s] %s(): ######  Time = %lds %ldms  ######\n",
					__FILE__, __FUNCTION__, tv.tv_sec, tv.tv_usec/1000);
#endif
}

void
list_exif_files (int portn, int busn, int devn, short force, uint32_t start, uint32_t num, int resfd)
{
	PTPParams params;
	PTP_USB ptp_usb;
	struct usb_device *dev = NULL;
	int i=0, max=0, errcode = -1;
	PTPObjectInfo oi;

	memset(&params, 0, sizeof(params));
	memset(&oi, 0, sizeof(oi));

	if (open_camera(portn, busn, devn, force, &ptp_usb, &params, &dev)<0){  /* [sawas-5] */
		printf("list_exif_files : Could not open_camera\n");
		close_camera(NULL,NULL,NULL);
		dev = NULL;
		goto out;
	}
	errcode = ptp_getdeviceinfo(&params, &params.deviceinfo);
	if(errcode != PTP_RC_OK){
		fprintf(stderr,"ERROR(%x): Could not get device info.\n", errcode);
		goto out;
	}
	errcode = ptp_getobjecthandles(&params, 0xffffffff,
								   PTP_OFC_EXIF_JPEG, 0x000000, &params.handles);
	if(errcode != PTP_RC_OK){
		fprintf(stderr,"ERROR(%x): Could not get object handles.\n", errcode);
		goto out;
	}
	if(start >= params.handles.n){
		errcode = -1;
		goto out;
	}
	if(num == 0) {
		max = params.handles.n;
	}
	else {
		max = start + num;
		if(max > params.handles.n){
			max = params.handles.n;
		}
	}
	for (i = start; i < max; i++) {
		memset(&oi, 0, sizeof(oi));
		errcode = ptp_getobjectinfo(&params,params.handles.Handler[i], &oi);
		if(errcode == PTP_RC_OK){
			print_object_info(i, params.handles.Handler[i], &oi, verbose);
		}
		if(resfd >= 0){
			write_result(resfd, i, errcode, params.handles.Handler[i], &oi);
		}
		if(errcode != PTP_RC_OK){
			goto out;
		}
		ptp_free_objectinfo(&oi);
	}

 out:
	if(errcode != PTP_RC_OK){
		// error result.
		fprintf(stderr,"ERROR(%x): list_exif_files.\n", errcode);
		if(resfd >= 0){
			memset(&oi, 0, sizeof(oi));
			write_result(resfd, i, errcode, 0, &oi);
		}
	}
	if(dev){
		ptp_free_deviceinfo(&params.deviceinfo);
		ptp_free_objecthandles(&params.handles);
		close_camera(&ptp_usb, &params, dev);
	}
	return;
}

void write_result_head(int fd, int action, uint32_t trid, uint32_t param1, uint32_t param2, uint32_t param3,
					   const char *model_name)  /* for Part-OBJ  [sawas-55] */
{
	char buf[BUF_SIZE];  /* for Part-OBJ  [sawas-55] */
	int ret,wlen;
	const char *WRITE_HEAD_FMT = "[ptpcam %c %u %08x %08x %08x] %s\n";  // for Part-OBJ  [sawas-55]

	memset(buf, 0, sizeof(buf));

	switch(action){
	case ACT_GET_DEVICE_INFO:
		snprintf(buf, sizeof(buf)-1, WRITE_HEAD_FMT, 'i', trid, param1, param2, param3, model_name);
		break;
	case ACT_GET_FILE:
		snprintf(buf, sizeof(buf)-1, WRITE_HEAD_FMT, 'g', trid, param1, param2, param3, model_name);
		break;
	case ACT_GET_PARTIAL_FILE:  /* for Part-OBJ  [sawas-55] */
		snprintf(buf, sizeof(buf)-1, WRITE_HEAD_FMT, 'P', trid, param1, param2, param3, model_name);
		break;
	case ACT_NUM_OF_OBJECTS:
		snprintf(buf, sizeof(buf)-1, WRITE_HEAD_FMT, 'n', trid, param1, param2, param3, model_name);
		break;
	case ACT_LIST_EXIF_FILES:
		snprintf(buf, sizeof(buf)-1, WRITE_HEAD_FMT, 'E', trid, param1, param2, param3, model_name);
		break;
	case ACT_GET_OBJECT_INFO:
		snprintf(buf, sizeof(buf)-1, WRITE_HEAD_FMT, 'I', trid, param1, param2, param3, model_name);
		break;
	case ACT_GET_THUMB_FILE:
		snprintf(buf, sizeof(buf)-1, WRITE_HEAD_FMT, 't', trid, param1, param2, param3, model_name);
		break;
	case ACT_FILE_OPER_QUIT:
		snprintf(buf, sizeof(buf)-1, WRITE_HEAD_FMT, 'q', trid, param1, param2, param3, model_name);
		break;
	default:
		break;
	}

	wlen = strlen(buf);
	
	ret = write(fd, buf, wlen);
	if(ret != wlen){
		printf("# ptpcam : Write result_head error(%d/%d)\n", ret, wlen);
	}
	return;
}

void write_result(int fd, int i, int errcode, uint32_t handle, const PTPObjectInfo *oi)
{
	/*							Num  Err  Handle Size Width Heigh Depth TSize TWidt THeig Cdate  Mdate  name  	   for Part-OBJ  [sawas-55] */
	const char *WRITE_OI_FMT = "N:%d E:%x H:%08x S:%d IW:%d IH:%d ID:%d TS:%d TW:%d TH:%d CD:%lu MD:%lu %s\n";  /* for Part-OBJ  [sawas-55] */
	char buf[BUF_SIZE];  /* for Part-OBJ  [sawas-55] */
	int ret,wlen;
	memset(buf, 0, sizeof(buf));

	snprintf(buf, sizeof(buf)-1, WRITE_OI_FMT, 
			 i, errcode, handle,
			 oi->ObjectCompressedSize,
			 oi->ImagePixWidth,
			 oi->ImagePixHeight,
			 oi->ImageBitDepth,
			 oi->ThumbCompressedSize,
			 oi->ThumbPixWidth,
			 oi->ThumbPixHeight,
			 (unsigned long)oi->CaptureDate,
			 (unsigned long)oi->ModificationDate,
			 oi->Filename
			 );
	
	wlen = strlen(buf);
	
	ret = write(fd, buf, wlen);
	if(ret != wlen){
		printf("# ptpcam : write result error(%d/%d)\n", ret, wlen);
	}
	return;
}

void write_result_string(int fd, int i, int errcode, const char *res_str)
{
	/*						   Number Error Strings  */
	const char *WRITE_STR_FMT = "N:%d E:%x %s\n";  /* for Part-OBJ  [sawas-55] */
	char buf[BUF_SIZE];  /* for Part-OBJ  [sawas-55] */
	int ret,wlen;

	memset(buf, 0, sizeof(buf));
	snprintf(buf, sizeof(buf)-1, WRITE_STR_FMT,
			 i, errcode, res_str);
	wlen = strlen(buf);
	
	ret = write(fd, buf, wlen);
	if(ret != wlen){
		printf("# ptpcam : write_result_string error(%d/%d)\n", ret, wlen);
	}
	return;
}
#endif

void
delete_object (int portn, int busn, int devn, short force, uint32_t handle)  /* [sawas-5] */
{
	PTPParams params;
	PTP_USB ptp_usb;
	struct usb_device *dev;
	PTPObjectInfo oi;
#if  1  /* ptpcam-memory-leak-fix  [SS-7] */
	int errcode = -1;
#endif

#if  1  /* ptpcam-data-initialization  [SS-6] */
	memset(&params, 0, sizeof(params));
	memset(&oi, 0, sizeof(oi));
#endif

#if  0  /* ptpcam-memory-leak-fix  [SS-7] */
	if (open_camera(busn, devn, force, &ptp_usb, &params, &dev)<0)
		return;
	CR(ptp_getobjectinfo(&params,handle,&oi),
		"Could not get object info\n");
	CR(ptp_deleteobject(&params, handle,0), "Could not delete object\n");
	printf("\nObject 0x%08x (%s) deleted.\n",handle, oi.Filename);
	close_camera(&ptp_usb, &params, dev);
#else
	if (open_camera(portn, busn, devn, force, &ptp_usb, &params, &dev)<0){  /* [sawas-5] */
		printf("delete_object : Could not open_camera\n");
		dev = NULL;
		goto out;
	}
	errcode = ptp_getobjectinfo(&params,handle, &oi);
	if(errcode != PTP_RC_OK){
		fprintf(stderr,"ERROR(%x): Could not get object info.\n", errcode);
		goto out;
	}
	errcode = ptp_deleteobject(&params, handle, 0);
	if(errcode != PTP_RC_OK){
		fprintf(stderr,"ERROR(%x): Could not delete object.\n", errcode);
		goto out;
	}
	printf("\nObject 0x%08x (%s) deleted.\n",handle, oi.Filename);
out:
	if(dev){
		ptp_free_objectinfo(&oi);
		close_camera(&ptp_usb, &params, dev);
	}
#endif
}

void
delete_all_files (int portn, int busn, int devn, short force)  /* [sawas-5] */
{
	PTPParams params;
	PTP_USB ptp_usb;
	struct usb_device *dev;
	PTPObjectInfo oi;
	uint32_t handle;
	int i;
	int ret;
#if  1  /* ptpcam-memory-leak-fix  [SS-7] */
	int errcode = -1;
#endif

#if  1  /* ptpcam-data-initialization  [SS-6] */
	memset(&params, 0, sizeof(params));
	memset(&oi, 0, sizeof(oi));
#endif

#if  0  /* ptpcam-memory-leak-fix  [SS-7] */
	if (open_camera(busn, devn, force, &ptp_usb, &params, &dev)<0)
		return;
	CR(ptp_getdeviceinfo (&params, &params.deviceinfo),
		"Could not get device info\n");
	printf("Camera: %s\n",params.deviceinfo.Model);
	CR(ptp_getobjecthandles (&params,0xffffffff, 0x000000, 0x000000,
		&params.handles),"Could not get object handles\n");
#else
	if (open_camera(portn, busn, devn, force, &ptp_usb, &params, &dev)<0){  /* [sawas-5] */
		printf("delete_all_files : Could not open_camera\n");
		dev = NULL;
		goto out;
	}
	errcode = ptp_getdeviceinfo(&params, &params.deviceinfo);
	if(errcode != PTP_RC_OK){
		fprintf(stderr,"ERROR(%x): Could not get device info.\n", errcode);
		goto out;
	}
	printf("Camera: %s\n",params.deviceinfo.Model);
	errcode = ptp_getobjecthandles (&params,0xffffffff, 0x000000, 0x000000,
		&params.handles);
	if(errcode != PTP_RC_OK){
		fprintf(stderr,"ERROR(%x): Could not get object handles\n", errcode);
		goto out;
	}
#endif

	for (i=0; i<params.handles.n; i++) {
		handle=params.handles.Handler[i];
#if  1  /* ptpcam-data-initialization  [SS-6] */
		memset(&oi, 0, sizeof(oi));
#endif
		if ((ret=ptp_getobjectinfo(&params,handle, &oi))!=PTP_RC_OK){
			fprintf(stderr,"Handle: 0x%08x\n",handle);
			fprintf(stderr,"ERROR: Could not get object info\n");
			ptp_perror(&params,ret);
			if (ret==PTP_ERROR_IO) clear_stall(&ptp_usb, dev);
			continue;
		}
#if  0  /* ptpcam-memory-leak-fix  [SS-7] */
		if (oi.ObjectFormat == PTP_OFC_Association)
			continue;
		CR(ptp_deleteobject(&params, handle,0),
				"Could not delete object\n");
		printf("Object 0x%08x (%s) deleted.\n",handle, oi.Filename);
	}
	close_camera(&ptp_usb, &params, dev);
#else
		if (oi.ObjectFormat == PTP_OFC_Association){
			ptp_free_objectinfo(&oi);
			continue;
		}
		errcode = ptp_deleteobject(&params, handle, 0);
		if(errcode != PTP_RC_OK){
			fprintf(stderr,"ERROR(%x): Could not delete object.\n", errcode);
			goto out;
		}
		printf("Object 0x%08x (%s) deleted.\n",handle, oi.Filename);
		ptp_free_objectinfo(&oi);
	}
out:
	if(dev){
		ptp_free_deviceinfo(&params.deviceinfo);
		ptp_free_objecthandles(&params.handles);
		close_camera(&ptp_usb, &params, dev);
	}
#endif
}

void
get_file (int portn, int busn, int devn, short force, uint32_t handle, char* filename,
		  int overwrite, int resfd)  /* [sawas-5] */
{
	PTPParams params;
	PTP_USB ptp_usb;
	struct usb_device *dev;
	struct utimbuf timebuf;
	int file;
	PTPObjectInfo oi;
	char *image;
	int ret;
#if  1  /* ptpcam-memory-leak-fix  [SS-7] */
	int errcode = -1;
#endif

#if  1  /* difference from original  [sawas-4] */
	char str_filename[256];
#ifdef  GET_PARTIAL_FILE_FOR_OL   /* [sawas-2.111] */
	uint32_t read_size;
#endif

	memset(str_filename, 0, sizeof(str_filename));
#endif

#if  1  /* ptpcam-data-initialization  [SS-6] */
	memset(&params, 0, sizeof(params));
	memset(&oi, 0, sizeof(oi));
#endif

#if  0  /* ptpcam-memory-leak-fix  [SS-7] */
	if (open_camera(busn, devn, force, &ptp_usb, &params, &dev)<0)
		return;
	CR(ptp_getdeviceinfo (&params, &params.deviceinfo),
		"Could not get device info\n");
	if (oi.ObjectFormat == PTP_OFC_Association)
			goto out;
	if (filename==NULL) filename=(oi.Filename);
	file=open(filename, (overwrite==OVERWRITE_EXISTING?0:O_EXCL)|O_RDWR|O_CREAT|O_TRUNC,S_IRWXU|S_IRGRP);
	if (file==-1) {
		if (errno==EEXIST) {
			printf("Skipping file: \"%s\", file exists!\n",filename);
			goto out;
		}
		perror("open");
		goto out;
	}
#else
	if (open_camera(portn, busn, devn, force, &ptp_usb, &params, &dev)<0){  /* [sawas-5] */
		printf("get_file : Could not open_camera\n");
		dev = NULL;
		goto out;
	}
	errcode = ptp_getdeviceinfo(&params, &params.deviceinfo);
	if(errcode != PTP_RC_OK){
		fprintf(stderr,"ERROR(%x): Could not get device info.\n", errcode);
		goto out;
	}
	printf("Camera: %s\n",params.deviceinfo.Model);

	if (verbose)
		printf ("Handle: 0x%08x\n",handle);
	errcode = ptp_getobjectinfo(&params,handle, &oi);
	if(errcode != PTP_RC_OK){
		fprintf(stderr,"ERROR(%x): Could not get object info.\n", errcode);
		goto out;
	}

#if  0  /* difference from original  [sawas-4] */
	if (oi.ObjectFormat == PTP_OFC_Association)
			goto out;
	if (filename==NULL) filename=(oi.Filename);
#else
	if(oi.ObjectFormat != PTP_OFC_EXIF_JPEG){
		errcode = -2;
		fprintf(stderr,"ERROR(%x): not EXIF_JPEG object.\n", errcode);
		goto out;
	}
	// --
	print_object_info(0, handle, &oi, verbose);
	// --
	if (filename==NULL) {
		snprintf(str_filename, sizeof(str_filename)-1, GET_FILE_DEFAULT_FMT,
				 handle, oi.Filename);
		filename=str_filename;
	}
#endif

	file=open(filename, (overwrite==OVERWRITE_EXISTING?0:O_EXCL)|O_RDWR|O_CREAT|O_TRUNC,S_IRWXU|S_IRGRP);
	if (file==-1) {
		if (errno==EEXIST) {
			printf("Skipping file: \"%s\", file exists!\n",filename);
			goto out;
		}
		perror("open");
		goto out;
	}
#endif

	lseek(file,oi.ObjectCompressedSize-1,SEEK_SET);
	write(file,"",1);
	if (file<0) goto out;
	image=mmap(0,oi.ObjectCompressedSize,PROT_READ|PROT_WRITE,MAP_SHARED,
		file,0);
	if (image==MAP_FAILED) {
		close(file);
		goto out;
	}
#if  0  /* difference from original  [sawas-4] */
	printf ("Saving file: \"%s\" ",filename);
#else
	printf ("[ptpcam] Saving file(map:%p,0x%x): \"%s\"\n",
		image, oi.ObjectCompressedSize, filename);
#endif
	fflush(NULL);

#ifdef  GET_PARTIAL_FILE_FOR_OL   /* [sawas-2.111] */
	if (is_gpo_supported(&params) &&
		(dev->descriptor.idVendor == VENDOR_ID_OL)) {
		printf("Vendor:0x%x --- use GPO\n", dev->descriptor.idVendor);
		ret = ptp_getpartialobject(&params,handle,&image,0,oi.ObjectCompressedSize,&read_size);  /* offset = 0 */
	}
	else {
		ret=ptp_getobject(&params,handle,&image);
	}
#else
	ret=ptp_getobject(&params,handle,&image);
#endif

	munmap(image,oi.ObjectCompressedSize);
	close(file);
	timebuf.actime=oi.ModificationDate;
	timebuf.modtime=oi.CaptureDate;
	utime(filename,&timebuf);
	if (ret!=PTP_RC_OK) {
		printf ("error!\n");
		ptp_perror(&params,ret);
	} else {
		printf("is done.\n");
	}
out:
#if  1  /* difference from original  [sawas-4] */
	if(resfd >= 0){
		write_result(resfd, 0, errcode, handle, &oi);
	}
#endif

#if  0  /* ptpcam-memory-leak-fix  [SS-7] */
	close_camera(&ptp_usb, &params, dev);
#else
	if(dev){
		ptp_free_deviceinfo(&params.deviceinfo);
		ptp_free_objectinfo(&oi);
		close_camera(&ptp_usb, &params, dev);
	}
	return;
#endif
}

#if  1  /* for Part-OBJ  [sawas-55] */
static PTPObjectInfo partial_oi;
static PTPParams partial_params;
static char partial_filename[256];

/**********************************
struct _PTPObjectInfo {
           ....
	char 	*Filename;  <--- 
           ....
	char	*Keywords;
};
**************************************/

uint32_t
get_partial_file (int portn, int busn, int devn, short force, uint32_t handle, char* filename,
				  uint32_t offset, uint32_t size, int overwrite, int resfd)
{
	/* PTPParams params; */
	PTP_USB ptp_usb;
	struct usb_device *dev;
	struct utimbuf timebuf;
	int file;
	/* PTPObjectInfo oi; */
	char *image;
	int errcode = -1;
	char str_filename[256];
	uint32_t readnum = 0;

	memset(str_filename, 0, sizeof(str_filename));

	/* ptpcam-data-initialization  [SS-6] */
	if (offset == 0) {
		memset(partial_filename, 0, sizeof(partial_filename));
		memset(&partial_params, 0, sizeof(partial_params));
		memset(&partial_oi, 0, sizeof(partial_oi));
	}
	else {
	  partial_oi.Filename = partial_filename;
	}

	/* check arguments */
	if (size == 0) {
		printf("get_partial_file : invalid size (0x%x)\n", size);
		dev = NULL;
		goto out;
	}

	if (open_camera(portn, busn, devn, force, &ptp_usb, &partial_params, &dev)<0){
		printf("get_partial_file : Could not open_camera\n");
		dev = NULL;
		goto out;
	}

	if (offset == 0) {
		/* get Device Info */
		errcode = ptp_getdeviceinfo(&partial_params, &partial_params.deviceinfo);
		if(errcode != PTP_RC_OK){
			fprintf(stderr,"ERROR(%x): Could not get device info.\n", errcode);
			goto out;
		}

		if (!is_gpo_supported(&partial_params)) {
			errcode = -1;
			fprintf(stderr,"[ptpcam] ERROR(%x): GetPartialObject unsupported.\n", errcode);
			close(file);
			goto out;
		}

		printf("Camera: %s\n",partial_params.deviceinfo.Model);
		if (verbose)
			printf ("Handle: 0x%08x\n",handle);

		/* get Object Info */
		errcode = ptp_getobjectinfo(&partial_params,handle, &partial_oi);
		if(errcode != PTP_RC_OK){
			fprintf(stderr,"ERROR(%x): Could not get object info.\n", errcode);
			goto out;
		}

		if(partial_oi.ObjectFormat != PTP_OFC_EXIF_JPEG){
			errcode = -2;
			fprintf(stderr,"ERROR(%x): not EXIF_JPEG object.\n", errcode);
			goto out;
		}
	}

	// --
	print_object_info(offset, handle, &partial_oi, verbose);
	// --

	if (filename==NULL) {
		snprintf(str_filename, sizeof(str_filename)-1, GET_FILE_DEFAULT_FMT,
			 	handle, partial_oi.Filename);
		filename=str_filename;
	}

	if (offset == 0) {
	    snprintf(partial_filename, sizeof(partial_filename), partial_oi.Filename);
	}

	file=open(filename, (overwrite==OVERWRITE_EXISTING?0:O_EXCL)|O_RDWR|O_CREAT|O_TRUNC,S_IRWXU|S_IRGRP);
	if (file==-1) {
		if (errno==EEXIST) {
			printf("Skipping file: \"%s\", file exists!\n",filename);
			goto out;
		}
		perror("open");
		goto out;
	}

	/* check size  [sawas2.11] */
	if ((offset + size) > partial_oi.ObjectCompressedSize) {
		size = partial_oi.ObjectCompressedSize - offset;
		/* printf("[ptpcam] Partial Size Over[0x%08x]!!!\n", size); */
	}

	lseek(file,offset+size-1,SEEK_SET);  /* [sawas2.11] */
	write(file,"",1);
	if (file<0) goto out;

	image=mmap(0,size,PROT_READ|PROT_WRITE,MAP_SHARED,file,0);
	if (image==MAP_FAILED) {
		close(file);
		goto out;
	}

	printf ("[ptpcam] Saving partial file(map:%p,0x%x): \"%s\"\n", image, size, filename);
	fflush(NULL);

	errcode=ptp_getpartialobject(&partial_params,handle,&image,offset,size,&readnum);
	munmap(image,size);
	close(file);

	/* time code */
	timebuf.actime=partial_oi.ModificationDate;
	timebuf.modtime=partial_oi.CaptureDate;
	utime(filename,&timebuf);

	if (errcode!=PTP_RC_OK) {
		printf ("error!\n");
		ptp_perror(&partial_params,errcode);
	} else {
		printf("is done.\n");
	}
out:
	if(resfd >= 0){
		write_result(resfd, readnum, errcode, handle, &partial_oi);
	}

	/* ptpcam-memory-leak-fix  [SS-7] */
	if(dev){
		ptp_free_deviceinfo(&partial_params.deviceinfo);

		if (offset == 0) {
		  ptp_free_objectinfo(&partial_oi);
		}

		close_camera(&ptp_usb, &partial_params, dev);
	}
	return readnum;
}
#endif

void
get_thumb_file (int portn, int busn, int devn, short force, uint32_t handle, char* filename,
				int overwrite, int resfd)  /* [sawas-5] */
{
	PTPParams params;
	PTP_USB ptp_usb;
	struct usb_device *dev = NULL;
	struct utimbuf timebuf;
	int file;
	PTPObjectInfo oi;
	char *image;
	int errcode = -1;
	char str_filename[256];

	memset(str_filename, 0, sizeof(str_filename));

#if  1  /* ptpcam-data-initialization  [SS-6] */
	memset(&params, 0, sizeof(params));
	memset(&oi, 0, sizeof(oi));
#endif
	if (open_camera(portn, busn, devn, force, &ptp_usb, &params, &dev)<0) {
		printf("get_thumb : Could not open_camera\n");
		dev = NULL;
		goto out;
	}
	errcode = ptp_getdeviceinfo(&params, &params.deviceinfo);
	if(errcode != PTP_RC_OK){
		fprintf(stderr,"ERROR(%x): Could not get device info.\n", errcode);
		goto out;
	}

	if (verbose) {
		printf ("GetThumb Handle: 0x%08x\n",handle);
		sleep(3);
	}
	errcode = ptp_getobjectinfo(&params,handle, &oi);
	if(errcode != PTP_RC_OK){
		fprintf(stderr,"ERROR(%x): Could not get object info.\n", errcode);
		goto out;
	}
	if(oi.ObjectFormat != PTP_OFC_EXIF_JPEG){
		errcode = -2;
		fprintf(stderr,"ERROR(%x): not EXIF_JPEG object.\n", errcode);
		goto out;
	}
	// --
	print_object_info(0, handle, &oi, verbose);
	// --
	if (filename==NULL) {
		snprintf(str_filename, sizeof(str_filename)-1, GET_THUMB_DEFAULT_FMT,
				 handle, oi.Filename);
		filename=str_filename;
	}
	file=open(filename, (overwrite==OVERWRITE_EXISTING?0:O_EXCL)|O_RDWR|O_CREAT|O_TRUNC,S_IRWXU|S_IRGRP);
	if (file==-1) {
		if (errno==EEXIST) {
			printf("[ptpcam] Skipping Thumb file: \"%s\", file exists!\n",filename);
			goto out;
		}
		perror("open");
		errcode = errno;
		goto out;
	}
	lseek(file,oi.ThumbCompressedSize-1,SEEK_SET);
	write(file,"",1);
	if (file<0) {
		errcode = errno;
		goto out;
	}
	image=mmap(0,oi.ThumbCompressedSize,PROT_READ|PROT_WRITE,MAP_SHARED,
			   file,0);
	if (image==MAP_FAILED) {
		errcode = errno;
		close(file);
		goto out;
	}
	printf ("[ptpcam] Saving Thumb file(map:%p,0x%x): \"%s\"\n",
			image, oi.ThumbCompressedSize, filename);
	// sleep(1);
	fflush(NULL);
	// printf ("[ptpcam] getthumb !!\n");
	errcode = ptp_getthumb(&params,handle,&image);
	// printf ("[ptpcam] getthumb .. done.\n");
	// sleep(1);
	munmap(image,oi.ThumbCompressedSize);
	close(file);
	timebuf.actime=oi.ModificationDate;
	timebuf.modtime=oi.CaptureDate;
	utime(filename,&timebuf);
	if (errcode != PTP_RC_OK) {
		printf ("error!\n");
		ptp_perror(&params, errcode);
	}
	// printf ("\n");

out:
	if(resfd >= 0){
		write_result(resfd, 0, errcode, handle, &oi);
	}
	if(dev){
		ptp_free_deviceinfo(&params.deviceinfo);
		ptp_free_objectinfo(&oi);
		close_camera(&ptp_usb, &params, dev);
	}
	return;
}

void
get_all_files (int portn, int busn, int devn, short force, int overwrite)  /* [sawas-5] */
{
	PTPParams params;
	PTP_USB ptp_usb;
	struct usb_device *dev;
	struct utimbuf timebuf;
	int file;
	PTPObjectInfo oi;
	uint32_t handle;
	char *image;
	int ret;
	int i;
	char *filename;
#if  1  /* ptpcam-memory-leak-fix  [SS-7] */
	int errcode = -1;
#endif

#if  1  /* ptpcam-data-initialization  [SS-6] */
	memset(&params, 0, sizeof(params));
	memset(&oi, 0, sizeof(oi));
#endif

#if  0  /* ptpcam-memory-leak-fix  [SS-7] */
	if (open_camera(busn, devn, force, &ptp_usb, &params, &dev)<0)
		return;
	CR(ptp_getdeviceinfo (&params, &params.deviceinfo),
		"Could not get device info\n");
	printf("Camera: %s\n",params.deviceinfo.Model);

	CR(ptp_getobjecthandles (&params,0xffffffff, 0x000000, 0x000000,
		&params.handles),"Could not get object handles\n");
#else
	if (open_camera(portn, busn, devn, force, &ptp_usb, &params, &dev)<0){  /* [sawas-5] */
		printf("get_all_files : Could not open_camera\n");
		dev = NULL;
		goto err;
	}
	errcode = ptp_getdeviceinfo(&params, &params.deviceinfo);
	if(errcode != PTP_RC_OK){
		fprintf(stderr,"ERROR(%x): Could not get device info.\n", errcode);
		goto err;
	}
	printf("Camera: %s\n",params.deviceinfo.Model);

	errcode = ptp_getobjecthandles (&params,0xffffffff, 0x000000, 0x000000,
		&params.handles);
	if(errcode != PTP_RC_OK){
		fprintf(stderr,"ERROR(%x): Could not get object handles\n", errcode);
		goto err;
	}
#endif

	for (i=0; i<params.handles.n; i++) {
		memset(&oi, 0, sizeof(PTPObjectInfo));
		handle=params.handles.Handler[i];
		if (verbose)
			printf ("Handle: 0x%08x\n",handle);
		if ((ret=ptp_getobjectinfo(&params,handle, &oi))!=PTP_RC_OK){
			fprintf(stderr,"ERROR: Could not get object info\n");
			ptp_perror(&params,ret);
			if (ret==PTP_ERROR_IO) clear_stall(&ptp_usb, dev);
			continue;
		}

		if (oi.ObjectFormat == PTP_OFC_Association)
				goto out;
		filename=(oi.Filename);
		file=open(filename, (overwrite==OVERWRITE_EXISTING?0:O_EXCL)|O_RDWR|O_CREAT|O_TRUNC,S_IRWXU|S_IRGRP);
		if (file==-1) {
			if (errno==EEXIST) {
				printf("Skipping file: \"%s\", file exists!\n",filename);
#if  0  /* ptpcam-memory-leak-fix  [SS-7] */
				continue;
#else
				goto out;
#endif
			}
			perror("open");
			goto out;
		}
		lseek(file,oi.ObjectCompressedSize-1,SEEK_SET);
		ret=write(file,"",1);
		if (ret==-1) {
			perror("write");
			goto out;
		}
		image=mmap(0,oi.ObjectCompressedSize,PROT_READ|PROT_WRITE,MAP_SHARED,
			file,0);
		if (image==MAP_FAILED) {
			perror("mmap");
			close(file);
			goto out;
		}
		printf ("Saving file: \"%s\" ",filename);
		fflush(NULL);
		ret=ptp_getobject(&params,handle,&image);
		munmap(image,oi.ObjectCompressedSize);
		close(file);
		timebuf.actime=oi.ModificationDate;
		timebuf.modtime=oi.CaptureDate;
		utime(filename,&timebuf);
		if (ret!=PTP_RC_OK) {
			printf ("error!\n");
			ptp_perror(&params,ret);
			if (ret==PTP_ERROR_IO) clear_stall(&ptp_usb, dev);
		} else {
			printf("is done.\n");
		}
out:
		ptp_free_objectinfo(&oi);
	}
err:
	if(dev){
		ptp_free_deviceinfo(&params.deviceinfo);
		ptp_free_objecthandles(&params.handles);
		close_camera(&ptp_usb, &params, dev);
	}
}


void
list_operations (int portn, int busn, int devn, short force)  /* [sawas-5] */
{
	PTPParams params;
	PTP_USB ptp_usb;
	struct usb_device *dev;
	int i;
	const char* name;
#if  1  /* ptpcam-memory-leak-fix  [SS-7] */
	int errcode = -1;
#endif

	printf("\nListing supported operations...\n");
#if 0
#ifdef DEBUG
	printf("%s: dev#%i bus#%i\n", __FUNCTION__, devn, busn);  /* [sawas-6] */
#endif
	dev=find_device(busn,devn,force);
	if (dev==NULL) {
		fprintf(stderr,"could not find any device matching given "
		"bus/dev numbers\n");
		exit(-1);
	}
	find_endpoints(dev,&ptp_usb.inep,&ptp_usb.outep,&ptp_usb.intep);

	init_ptp_usb(&params, &ptp_usb, dev);
	CR(ptp_opensession (&params,1),
		"Could not open session!\n");
#endif

#if  1  /* ptpcam-data-initialization  [SS-6] */
	memset(&params, 0, sizeof(params));
#endif

#if  0  /* ptpcam-memory-leak-fix  [SS-7] */
	if (open_camera(busn, devn, force, &ptp_usb, &params, &dev)<0)
		return;
	CR(ptp_getdeviceinfo (&params, &params.deviceinfo),
		"Could not get device info\n");
#else
	if (open_camera(portn, busn, devn, force, &ptp_usb, &params, &dev)<0){  /* [sawas-5] */
		printf("list_operations : Could not open_camera\n");
		dev = NULL;
		goto out;
	}
	errcode = ptp_getdeviceinfo(&params, &params.deviceinfo);
	if(errcode != PTP_RC_OK){
		fprintf(stderr,"ERROR(%x): Could not get device info.\n", errcode);
		goto out;
	}
#endif

	printf("Camera: %s\n",params.deviceinfo.Model);
	for (i=0; i<params.deviceinfo.OperationsSupported_len; i++)
	{
		name=ptp_get_operation_name(&params,
			params.deviceinfo.OperationsSupported[i]);

		if (name==NULL)
			printf("  0x%04x: 0x%04x\n",
				params.deviceinfo.OperationsSupported[i],
				params.deviceinfo.OperationsSupported[i]);
		else
			printf("  0x%04x: %s\n",
				params.deviceinfo.OperationsSupported[i],name);
	}
#if  0  /* ptpcam-memory-leak-fix  [SS-7] */
	CR(ptp_closesession(&params), "Could not close session!\n");
	close_usb(&ptp_usb, dev);
#else
out:
	if(dev){
		ptp_free_deviceinfo(&params.deviceinfo);
		close_camera(&ptp_usb, &params, dev);
	}
#endif
}

void
list_properties (int portn, int busn, int devn, short force)  /* [sawas-5] */
{
	PTPParams params;
	PTP_USB ptp_usb;
	struct usb_device *dev;
	const char* propdesc;
	int i;
#if  1  /* ptpcam-memory-leak-fix  [SS-7] */
	int errcode = -1;
#endif

#if  1  /* ptpcam-data-initialization  [SS-6] */
	memset(&params, 0, sizeof(params));
#endif
	printf("\nListing properties...\n");
#ifdef DEBUG
	printf("%s: dev#%i bus#%i\n", __FUNCTION__, devn, busn);  /* [sawas-6] */
#endif
#if 0
	dev=find_device(busn,devn,force);
	if (dev==NULL) {
		fprintf(stderr,"could not find any device matching given "
		"bus/dev numbers\n");
		exit(-1);
	}
	find_endpoints(dev,&ptp_usb.inep,&ptp_usb.outep,&ptp_usb.intep);

	init_ptp_usb(&params, &ptp_usb, dev);
	CR(ptp_opensession (&params,1),
		"Could not open session!\n");
#endif

#if  0  /* ptpcam-memory-leak-fix  [SS-7] */
	if (open_camera(busn, devn, force, &ptp_usb, &params, &dev)<0)
		return;
	CR(ptp_getdeviceinfo (&params, &params.deviceinfo),
		"Could not get device info\n");
#else
	if (open_camera(portn, busn, devn, force, &ptp_usb, &params, &dev)<0){  /* [sawas-5] */
		printf("list_properties : Could not open_camera\n");
		dev = NULL;
		goto out;
	}
	errcode = ptp_getdeviceinfo(&params, &params.deviceinfo);
	if(errcode != PTP_RC_OK){
		fprintf(stderr,"ERROR(%x): Could not get device info.\n", errcode);
		goto out;
	}
#endif

	printf("Camera: %s\n",params.deviceinfo.Model);
	for (i=0; i<params.deviceinfo.DevicePropertiesSupported_len;i++){
		propdesc=ptp_get_property_name(&params,
			params.deviceinfo.DevicePropertiesSupported[i]);
		if (propdesc!=NULL) 
			printf("  0x%04x: %s\n",
				params.deviceinfo.DevicePropertiesSupported[i],
				propdesc);
		else
			printf("  0x%04x: 0x%04x\n",
				params.deviceinfo.DevicePropertiesSupported[i],
				params.deviceinfo.DevicePropertiesSupported[i]);
	}
#if  0  /* ptpcam-memory-leak-fix  [SS-7] */
	CR(ptp_closesession(&params), "Could not close session!\n");
	close_usb(&ptp_usb, dev);
#else
out:
	if(dev){
		ptp_free_deviceinfo(&params.deviceinfo);
		close_camera(&ptp_usb, &params, dev);
	}
#endif
}

short
print_propval (uint16_t datatype, void* value, short hex);
short
print_propval (uint16_t datatype, void* value, short hex)
{
	switch (datatype) {
		case PTP_DTC_INT8:
			printf("%hhi",*(char*)value);
			return 0;
		case PTP_DTC_UINT8:
			printf("%hhu",*(unsigned char*)value);
			return 0;
		case PTP_DTC_INT16:
			printf("%hi",*(int16_t*)value);
			return 0;
		case PTP_DTC_UINT16:
			if (hex==PTPCAM_PRINT_HEX)
				printf("0x%04hX (%hi)",*(uint16_t*)value,
					*(uint16_t*)value);
			else
				printf("%hi",*(uint16_t*)value);
			return 0;
		case PTP_DTC_INT32:
			printf("%i",*(int32_t*)value);
			return 0;
		case PTP_DTC_UINT32:
			if (hex==PTPCAM_PRINT_HEX)
				printf("0x%08X (%i)",*(uint32_t*)value,
					*(uint32_t*)value);
			else
				printf("%i",*(uint32_t*)value);
			return 0;
		case PTP_DTC_STR:
			printf("\"%s\"",(char *)value);
	}
	return -1;
}

uint16_t
set_property (PTPParams* params,
		uint16_t property, char* value, uint16_t datatype);
uint16_t
set_property (PTPParams* params,
		uint16_t property, char* value, uint16_t datatype)
{
	void* val=NULL;
#if  1  /* ptpcam-free-value  [SS-4] */
	uint16_t ret = 0;
#endif

	switch(datatype) {
	case PTP_DTC_INT8:
		val=malloc(sizeof(int8_t));
		*(int8_t*)val=(int8_t)strtol(value,NULL,0);
		break;
	case PTP_DTC_UINT8:
		val=malloc(sizeof(uint8_t));
		*(uint8_t*)val=(uint8_t)strtol(value,NULL,0);
		break;
	case PTP_DTC_INT16:
		val=malloc(sizeof(int16_t));
		*(int16_t*)val=(int16_t)strtol(value,NULL,0);
		break;
	case PTP_DTC_UINT16:
		val=malloc(sizeof(uint16_t));
		*(uint16_t*)val=(uint16_t)strtol(value,NULL,0);
		break;
	case PTP_DTC_INT32:
		val=malloc(sizeof(int32_t));
		*(int32_t*)val=(int32_t)strtol(value,NULL,0);
		break;
	case PTP_DTC_UINT32:
		val=malloc(sizeof(uint32_t));
		*(uint32_t*)val=(uint32_t)strtol(value,NULL,0);
		break;
	case PTP_DTC_STR:
		val=(void *)value;
	}
#if  0  /* ptpcam-free-value  [SS-4] */
	return(ptp_setdevicepropvalue(params, property, val, datatype));
	free(val);
	return 0;
#else
	ret = ptp_setdevicepropvalue(params, property, val, datatype);
	free(val);
	return ret;
#endif
}

void
getset_property (int portn, int busn,int devn,uint16_t property,char* value,short force);  /* [sawas-5] */
void
getset_property (int portn, int busn,int devn,uint16_t property,char* value,short force)  /* [sawas-5] */
{
	PTPParams params;
	PTP_USB ptp_usb;
	struct usb_device *dev;
	PTPDevicePropDesc dpd;
	const char* propdesc;

#if  1  /* ptpcam-data-initialization  [SS-6] */
	memset(&params, 0, sizeof(params));
#endif
	printf ("\n");
#ifdef DEBUG
	printf("%s: dev#%i bus#%i\n", __FUNCTION__, devn, busn);  /* [sawas-6] */
#endif

	dev=find_device(portn,busn,devn,force);  /* [sawas-5] */
	if (dev==NULL) {
		fprintf(stderr,"could not find any device matching given "
		"bus/dev numbers\n");
		exit(-1);
	}
	find_endpoints(dev,&ptp_usb.inep,&ptp_usb.outep,&ptp_usb.intep);

	init_ptp_usb(&params, &ptp_usb, dev);
	CR(ptp_opensession (&params,1),
		"Could not open session!\nTry to reset the camera.\n");
	CR(ptp_getdeviceinfo (&params, &params.deviceinfo),
		"Could not get device info\nTry to reset the camera.\n");
	propdesc=ptp_get_property_name(&params,property);
	printf("Camera: %s",params.deviceinfo.Model);
	if ((devn!=0)||(busn!=0)) 
		printf(" (bus %i, dev %i)\n",busn,devn);
	else
		printf("\n");
	if (!ptp_property_issupported(&params, property))
	{
		fprintf(stderr,"The device does not support this property!\n");
		CR(ptp_closesession(&params), "Could not close session!\n"
			"Try to reset the camera.\n");
		return;
	}
	printf("Property '%s'\n",propdesc);
	memset(&dpd,0,sizeof(dpd));
	CR(ptp_getdevicepropdesc(&params,property,&dpd),
		"Could not get device property description!\n"
		"Try to reset the camera.\n");
	if (verbose)
		printf ("Data type is 0x%04x\n",dpd.DataType);
	printf ("Current value is ");
	if (dpd.FormFlag==PTP_DPFF_Enumeration)
		PRINT_PROPVAL_DEC(dpd.CurrentValue);
	else 
		PRINT_PROPVAL_HEX(dpd.CurrentValue);
	printf("\n");

	if (value==NULL) {
		printf ("Factory default value is ");
		if (dpd.FormFlag==PTP_DPFF_Enumeration)
			PRINT_PROPVAL_DEC(dpd.FactoryDefaultValue);
		else 
			PRINT_PROPVAL_HEX(dpd.FactoryDefaultValue);
		printf("\n");
		printf("The property is ");
		if (dpd.GetSet==PTP_DPGS_Get)
			printf ("read only");
		else
			printf ("settable");
		switch (dpd.FormFlag) {
		case PTP_DPFF_Enumeration:
			printf (", enumerated. Allowed values are:\n");
			{
				int i;
				for(i=0;i<dpd.FORM.Enum.NumberOfValues;i++){
					PRINT_PROPVAL_HEX(
					dpd.FORM.Enum.SupportedValue[i]);
					printf("\n");
				}
			}
			break;
		case PTP_DPFF_Range:
			printf (", within range:\n");
			PRINT_PROPVAL_DEC(dpd.FORM.Range.MinimumValue);
			printf(" - ");
			PRINT_PROPVAL_DEC(dpd.FORM.Range.MaximumValue);
			printf("; step size: ");
			PRINT_PROPVAL_DEC(dpd.FORM.Range.StepSize);
			printf("\n");
			break;
		case PTP_DPFF_None:
			printf(".\n");
		}
	} else {
		uint16_t r;
		printf("Setting property value to '%s'\n",value);
		r=(set_property(&params, property, value, dpd.DataType));
		if (r!=PTP_RC_OK)
				ptp_perror(&params,r);
	}
	ptp_free_devicepropdesc(&dpd);
#if  1  /* ptpcam-free-value  [SS-4] */
	ptp_free_deviceinfo(&params.deviceinfo);
#endif
	CR(ptp_closesession(&params), "Could not close session!\n"
	"Try to reset the camera.\n");
	close_usb(&ptp_usb, dev);
}

int
usb_get_endpoint_status(PTP_USB* ptp_usb, int ep, uint16_t* status)
{
	 return (usb_control_msg(ptp_usb->handle,
		USB_DP_DTH|USB_RECIP_ENDPOINT, USB_REQ_GET_STATUS,
		USB_FEATURE_HALT, ep, (char *)status, 2, 3000));
}

int
usb_clear_stall_feature(PTP_USB* ptp_usb, int ep)
{

	return (usb_control_msg(ptp_usb->handle,
		USB_RECIP_ENDPOINT, USB_REQ_CLEAR_FEATURE, USB_FEATURE_HALT,
		ep, NULL, 0, 3000));
}

int
usb_ptp_get_device_status(PTP_USB* ptp_usb, uint16_t* devstatus);
int
usb_ptp_get_device_status(PTP_USB* ptp_usb, uint16_t* devstatus)
{
	return (usb_control_msg(ptp_usb->handle,
		USB_DP_DTH|USB_TYPE_CLASS|USB_RECIP_INTERFACE,
		USB_REQ_GET_DEVICE_STATUS, 0, 0,
		(char *)devstatus, 4, 3000));
}

int
usb_ptp_device_reset(PTP_USB* ptp_usb);
int
usb_ptp_device_reset(PTP_USB* ptp_usb)
{
	return (usb_control_msg(ptp_usb->handle,
		USB_TYPE_CLASS|USB_RECIP_INTERFACE,
		USB_REQ_DEVICE_RESET, 0, 0, NULL, 0, 3000));
}

void
reset_device (int portn, int busn, int devn, short force);  /* [sawas-5] */
void
reset_device (int portn, int busn, int devn, short force)  /* [sawas-5] */
{
	PTPParams params;
	PTP_USB ptp_usb;
	struct usb_device *dev;
	uint16_t status;
	uint16_t devstatus[2] = {0,0};
	int ret;

#if  1  /* ptpcam-data-initialization  [SS-6] */
	memset(&params, 0, sizeof(params));
#endif

#ifdef DEBUG
	printf("%s: dev#%i bus#%i\n", __FUNCTION__, devn, busn);  /* [sawas-6] */
#endif

	dev=find_device(portn,busn,devn,force);  /* [sawas-5] */
	if (dev==NULL) {
		fprintf(stderr,"could not find any device matching given "
		"bus/dev numbers\n");
		exit(-1);
	}
	find_endpoints(dev,&ptp_usb.inep,&ptp_usb.outep,&ptp_usb.intep);

	init_ptp_usb(&params, &ptp_usb, dev);
	
	/* get device status (devices likes that regardless of its result)*/
	usb_ptp_get_device_status(&ptp_usb,devstatus);
	
	/* check the in endpoint status*/
	ret = usb_get_endpoint_status(&ptp_usb,ptp_usb.inep,&status);
	if (ret<0) perror ("usb_get_endpoint_status()");
	/* and clear the HALT condition if happend*/
	if (status) {
		printf("Resetting input pipe!\n");
		ret=usb_clear_stall_feature(&ptp_usb,ptp_usb.inep);
		if (ret<0)perror ("usb_clear_stall_feature()");
	}
	status=0;
	/* check the out endpoint status*/
	ret = usb_get_endpoint_status(&ptp_usb,ptp_usb.outep,&status);
	if (ret<0) perror ("usb_get_endpoint_status()");
	/* and clear the HALT condition if happend*/
	if (status) {
		printf("Resetting output pipe!\n");
		ret=usb_clear_stall_feature(&ptp_usb,ptp_usb.outep);
		if (ret<0)perror ("usb_clear_stall_feature()");
	}
	status=0;
	/* check the interrupt endpoint status*/
	ret = usb_get_endpoint_status(&ptp_usb,ptp_usb.intep,&status);
	if (ret<0)perror ("usb_get_endpoint_status()");
	/* and clear the HALT condition if happend*/
	if (status) {
		printf ("Resetting interrupt pipe!\n");
		ret=usb_clear_stall_feature(&ptp_usb,ptp_usb.intep);
		if (ret<0)perror ("usb_clear_stall_feature()");
	}

	/* get device status (now there should be some results)*/
	ret = usb_ptp_get_device_status(&ptp_usb,devstatus);
	if (ret<0) 
		perror ("usb_ptp_get_device_status()");
	else	{
		if (devstatus[1]==PTP_RC_OK) 
			printf ("Device status OK\n");
		else
			printf ("Device status 0x%04x\n",devstatus[1]);
	}
	
	/* finally reset the device (that clears prevoiusly opened sessions)*/
	ret = usb_ptp_device_reset(&ptp_usb);
	if (ret<0)perror ("usb_ptp_device_reset()");
	/* get device status (devices likes that regardless of its result)*/
	usb_ptp_get_device_status(&ptp_usb,devstatus);

	close_usb(&ptp_usb, dev);

}

#if  1  /* difference from original  [sawas-4] */
void print_object_info(int n, uint32_t handle, const PTPObjectInfo *oi, int vflag)
{
#define CPL " "
	if(!vflag) printf("[OI]");
	printf(CPL"%05i", n);

	if(vflag) printf("\nHandle\t\t:");
	printf(CPL"0x%08x", handle);
#if 0
	if(vflag) printf("\nStorageID\t:");
	printf(CPL"%08x", oi->StorageID);
#endif

	if(vflag) printf("\nObjectFormat\t:");
	printf(CPL"0x%04x", oi->ObjectFormat);

#if 0
	if(vflag) printf("\nProtectionStat\t:");
	printf(CPL"0x%04x", oi->ProtectionStatus);
#endif

	if(vflag) printf("\nObjectSize\t:");
	printf(CPL"%i", oi->ObjectCompressedSize);
	if(vflag) printf("\nImagePixWidth\t:");
	printf(CPL"%d", oi->ImagePixWidth);
	if(vflag) printf("\nImagePixHeight\t:");
	printf(CPL"%d", oi->ImagePixHeight);
	if(vflag) printf("\nImageBitDepth\t:");
	printf(CPL"%d", oi->ImageBitDepth);

	if(vflag) printf("\nThumbSize\t:");
	printf(CPL"%i", oi->ThumbCompressedSize);
	if(vflag) printf("\nThumbPixWidth\t:");
	printf(CPL"%d", oi->ThumbPixWidth);
	if(vflag) printf("\nThumbPixHeight\t:");
	printf(CPL"%d", oi->ThumbPixHeight);

#if 0
	if(vflag) printf("\nAssociationType\t:");
	printf(CPL"%d", oi->AssociationType);
	if(vflag) printf("\nAssociationDesc\t:");
	printf(CPL"%d", oi->AssociationDesc);

	if(vflag) printf("\nSequenceNum\t:");
	printf(CPL"%d", oi->SequenceNumber);
#endif

	if(vflag) printf("\nFileName\t:");
	printf(CPL"%s", oi->Filename);

	if(vflag) printf("\nCaptureDate\t:");
	printf(CPL"%d", (int)oi->CaptureDate);
	if(vflag) printf(CPL"%s",ctime(&oi->CaptureDate));

	if(vflag) printf("\nModifyDate\t:");
	printf(CPL"%d", (int)oi->ModificationDate);
	if(vflag) printf(CPL"%s", ctime(&oi->ModificationDate));

#if 0
	if(vflag) printf("\nKeywords\t:");
	if(vflag) printf(CPL"%s", oi->Keywords);
#endif

	printf("\n");
}

void
get_object_info (int portn, int busn, int devn, short force, uint32_t handle, int resfd)
{
	PTPParams params;
	PTP_USB ptp_usb;
	struct usb_device *dev = NULL;
	PTPObjectInfo oi;
	int errcode = -1;
	memset(&params, 0, sizeof(params));
	memset(&oi, 0, sizeof(oi));

	if (open_camera(portn, busn, devn, force, &ptp_usb, &params, &dev)<0){  /* [sawas-5] */
		printf("get_object_info : Could not open_camera\n");
		close_camera(NULL,NULL,NULL);
		dev = NULL;
		goto out;
	}

	errcode = ptp_getdeviceinfo(&params, &params.deviceinfo);
	if(errcode != PTP_RC_OK){
		fprintf(stderr,"ERROR(%x): Could not get device info.\n", errcode);
		goto out;
	}

	if (verbose)
		printf ("Handle: 0x%08x\n",handle);
	errcode = ptp_getobjectinfo(&params,handle, &oi);
	if(resfd >= 0){
		write_result(resfd, 0, errcode, handle, &oi);
	}
	if(errcode != PTP_RC_OK){
		goto out;
	}
	if (verbose) {
		print_object_info(0, handle, &oi, verbose);
		printf("\n");
	}

 out:
	if(errcode != PTP_RC_OK){
		// error result.
		fprintf(stderr,"ERROR(%x): get_object_info.\n", errcode);
		if(resfd >= 0){
			write_result(resfd, 0, errcode, handle, &oi);
		}
	}

	if(dev){
		ptp_free_deviceinfo(&params.deviceinfo);
		ptp_free_objectinfo(&oi);
		close_camera(&ptp_usb, &params, dev);
	}
	return;
}

/* GPO support check for Part-OBJ  [sawas-55] */
int is_gpo_supported(PTPParams *params)
{
	int i, gpo_supported = 0;

	for (i=0; i<params->deviceinfo.OperationsSupported_len; i++)
	{
		if (params->deviceinfo.OperationsSupported[i] == PTP_OC_GetPartialObject) {
			gpo_supported = 1;
#if  0  /* 1: for debug  [sawas-55] */
			printf("  0x%04x\n", params.deviceinfo.OperationsSupported[i]);
			printf("GPO supported\n");
#endif
		}
	}

	return gpo_supported;
}

int get_info (int portn, int busn, int devn, short force,
			  char *model, int model_size, char *manuf, int manuf_size, int *gpo_supported)  /* for Part-OBJ  [sawas-55] */
{
	PTPParams params;
	PTP_USB ptp_usb;
	struct usb_device *dev = NULL;
	int errcode = -1;  /* [sawas-55] */

	memset(&params, 0, sizeof(params));
	if (open_camera(portn, busn, devn, force, &ptp_usb, &params, &dev)<0){  /* [sawas-5] */
		// printf("getinfo : Could not open_camera(%p)\n",dev);
		close_camera(NULL,NULL,NULL);
		dev = NULL;
		goto out;
	}

#if  0  /* 1: for debug  [sawas-5] */
	printf("[%s:%s]open_camera() done. {port=%d}\n", __FILE__, __FUNCTION__, portn);
#endif

	errcode = ptp_getdeviceinfo(&params, &params.deviceinfo);
	if(errcode != PTP_RC_OK){
		printf("getinfo : Could not get device info\n");
		goto out;
	}

#if 0
	printf("getinfo : %s, %s\n",
		   params.deviceinfo.Manufacturer,
		   params.deviceinfo.Model );
#endif

#if  0  /* NULL Access to dead  [sawas-4] */
	strncpy(model, params.deviceinfo.Model, model_size);
   	model[model_size-1] = '\0';
	strncpy(manuf, params.deviceinfo.Manufacturer, manuf_size);
	manuf[manuf_size-1] = '\0';
#else

#if  0  /* 1: for debug  [sawas-4] */
	printf("[%s:%s] ptp_getdeviceinfo() done.\n", __FILE__, __FUNCTION__);

	printf("model = '%s'\n", model);
	printf("MODEL = '%s'\n", params.deviceinfo.Model);
	printf("model_size = %d\n", model_size);

	printf("\n");
	printf("manuf = '%s'\n", manuf);
	printf("Manufacturer = '%s'\n", params.deviceinfo.Manufacturer);
	printf("manuf_size = %d\n", manuf_size);
	sleep(1);
#endif

	if (params.deviceinfo.Model) {
		strncpy(model, params.deviceinfo.Model, model_size);
		model[model_size-1] = '\0';
	}
	else {
		model[0] = '\0';
	}

	if (params.deviceinfo.Manufacturer) {
		strncpy(manuf, params.deviceinfo.Manufacturer, manuf_size);
		manuf[manuf_size-1] = '\0';
	}
	else {
		manuf[0] = '\0';
	}
#endif

	/* GPO support check  [sawas-55] */
	*gpo_supported = is_gpo_supported(&params);

 out:
	if(dev){
		ptp_free_deviceinfo(&params.deviceinfo);
		close_camera(&ptp_usb, &params, dev);
	}

	return errcode;
}

void file_operation (int portn, int busn, int devn, short force, char* execdir)
{
	int action=0, loop_cnt=0;
	int ret=0, cmdfd=-1, resfd=-1, lckfd=-1;
	char cmdbuf[BUF_SIZE], cmd_option = '\0';  /* for Part-OBJ  [sawas-55] */
	uint32_t trid=0, param1=0, param2=0, handle=0, ofc=(uint32_t)-1, start=0, num=0;
	int size=0;  /* [sawas-55] */
	uint32_t param3=0;  /* [sawas-55] */
	int gpo_supported=0;  /* [sawas-55] */
	int overwrite=SKIP_IF_EXISTS, b_exit = 0;
	char *filename=NULL;
	char model_name[32], manuf_name[32];
	int  i, loop, offset;  /* [sawas-55] */

	printf("# ptpcam ver.%s (pid=%d) FILE_OPERATION : %s\n",
							PTPCAM_VERSION, getpid(), execdir);

	if( chdir(execdir) ){
		printf("# ptpcam : Error FILE_OPERATION - execdir.\n");
		return;
	}
	
	loop_cnt = 0;

	while(1){
#ifdef  DEBUG_POLL  // [sawas-5]
		if(loop_cnt%5 == 0){
#else
		if(loop_cnt%30 == 0){
#endif
			loop = 1;
			if (hub_id) {  // [sawas-55]
				loop++;
			}

			if (use_ohci2_port) {  // OHCI  [sawas-6]
				loop++;
			}

			for (i = 0; i < loop; i++) {  // [sawas-55]
				if (hub_id) {  // [sawas-6]
					hub_port = i;  // 0/1  [sawas-55]
				}

				if (use_ohci2_port) {  // OHCI  [sawas-6]
					ohci2_port = i;  // 0/1
				}

				// all hardware port under hub
				for (portn = 0; portn < scan_port; portn++) {  // [sawas-55]
					// set hub port
					memset(model_name, 0, sizeof(model_name));
					memset(manuf_name, 0, sizeof(manuf_name));
					ret = get_info(portn, busn, devn, force,
								   model_name, sizeof(model_name),
								   manuf_name, sizeof(manuf_name), &gpo_supported);  /* [sawas-55] */
#ifdef  DEBUG_POLL  // [sawas-6]
					if(ret == PTP_RC_OK){  // [sawas-5]
						printf("#ptpcam[%d] device_info [%s : %s] {portn: %x bus:%d dev:%d%s%s}\n",
							loop_cnt, manuf_name, model_name, portn, busn, devn,
							hub_port   ? " hub"   : "",
							ohci2_port ? " OHCI2" : "");  // OHCI  [sawas-6]
					}
					else {
						printf("#ptpcam[%d] device_info error. {portn:%x bus:%d dev:%d%s%s}\n",
									loop_cnt, portn, busn, devn,
									hub_port   ? " hub"   : "",
									ohci2_port ? " OHCI2" : "");  // OHCI  [sawas-6]
					}
#endif
				}
			}
		}

		loop_cnt++;
		sleep(1);

		if( access(FILE_OPERATION_CMDFILE, R_OK) == 0){
			//---------------------------------------------------------------------
			// Command.
			//---------------------------------------------------------------------
			lckfd = open(FILE_OPERATION_CMDLOCK, O_CREAT | O_EXCL, 0644);
			if(lckfd < 0){
				if(errno == EEXIST){
					printf("# ptpcam : FILE_OPERATION - already cmd locked.\n");
				}
				else {
					printf("# ptpcam : FILE_OPERATION - cmd lock error(%d).\n",errno);
				}
				continue;
			}
			// CMD Lock OK.
			close(lckfd);

			cmdfd = open(FILE_OPERATION_CMDFILE, O_RDONLY);
			if(cmdfd < 0){
				printf("# ptpcam : FILE_OPERATION - CMD open err(%s).\n", FILE_OPERATION_CMDFILE);
				goto CMD_END;
			}
			memset(cmdbuf, 0, sizeof(cmdbuf));
			ret = read(cmdfd, cmdbuf, sizeof(cmdbuf));
			if(ret <= 0){
				printf("# ptpcam : FILE_OPERATION - CMD read err(%s).\n", FILE_OPERATION_CMDFILE);
				close(cmdfd);
				goto CMD_END;
			}

			// Parse CMDFILE.
			ret = sscanf(cmdbuf, "ptpcam %c %u %02x %08x %08x %08x",  /* OHCI  [sawas-6] */
						&cmd_option, &trid, &portn, &param1, &param2, &param3);  /* for Part-OBJ  [sawas-55] */
			printf("# ptpcam : FILE_OP - CMD sscanf(%d) [op:%c, tr:%u port:0x%02x p1:0x%08x p2:0x%08x p3:0x%08x]\n",  /* [sawas-6] */
				   ret, cmd_option, trid, portn, param1, param2, param3);

			/* get hub port  [sawas-55] */
			hub_port =   portn & 0x80;  /* bit7  [sawas-6] */

			/* get OHCI#2 port  [sawas-6] */
			ohci2_port = portn & 0x20;  /* bit5 OHCI  [sawas-6] */

			/* get port  [sawas-6] */
			portn &= 0xF;  /* lower 4bits  [sawas-6] */

#ifdef  DEBUG_SEARCH
			printf("hub_port  =0x%02x\n", hub_port  );
			printf("ohci2_port=0x%02x\n", ohci2_port);  /* OHCI  [sawas-6] */
			printf("port      =0x%02x\n", portn     );  /* [sawas-6] */
#endif

			switch(cmd_option){
			case 'i':
				action = ACT_GET_DEVICE_INFO;
				break;
			case 'n':
				action = ACT_NUM_OF_OBJECTS;
				ofc = param1;
				break;
			case 'I':
				action = ACT_GET_OBJECT_INFO;
				handle = param1;
				break;
			case 'E':
				action = ACT_LIST_EXIF_FILES;
				start = param1;
				num = param2;
				break;
			case 'g':
				action = ACT_GET_FILE;
				handle = param1;
				break;
#if  1  /* for Part-OBJ  [sawas-55] */
			case 'P':
				action = ACT_GET_PARTIAL_FILE;
				handle = param1;
				offset = param2;
				size   = param3;
				break;
#endif
			case 't':
				action = ACT_GET_THUMB_FILE;
				handle = param1;
				break;
			case 'q':
				action = ACT_FILE_OPER_QUIT;
				break;
			default:
				action = ACT_NONE;
				break;
			}

			close(cmdfd);

			// delete CMDFILE.
			// unlink(FILE_OPERATION_CMDFILE);
			rename(FILE_OPERATION_CMDFILE, FILE_OPERATION_CMDLAST);


			//---------------------------------------------------------------------
			// Response.
			//---------------------------------------------------------------------
			lckfd = open(FILE_OPERATION_RESLOCK, O_CREAT | O_EXCL, 0644);
			if(lckfd < 0){
				if(errno == EEXIST){
					printf("# ptpcam : FILE_OPERATION - already res locked.\n");
				}
				else {
					printf("# ptpcam : FILE_OPERATION - res lock error(%d).\n",errno);
				}
				goto CMD_END;
			}
			// RES Lock OK.
			close(lckfd);

			unlink(FILE_OPERATION_RESFILE);
			resfd = open(FILE_OPERATION_RESFILE, O_CREAT|O_WRONLY|O_EXCL, 0644);
			if(resfd < 0){
				goto RES_END;
			}

			// get device info.
			ret = get_info(portn, busn, devn, force,
						   model_name, sizeof(model_name),
						   manuf_name, sizeof(manuf_name), &gpo_supported);  /* for Part-OBJ  [sawas-55] */
			// write head
			write_result_head(resfd, action, trid, param1, param2, param3, model_name);  /* for Part-OBJ  [sawas-55] */

			if(ret != PTP_RC_OK){
				//printf("# ptpcam : loop - get_info error(%x).\n",ret);  for Part-OBJ  [sawas-55]
				// error occurred.

				memset(model_name, 0, sizeof(model_name));
				memset(manuf_name, 0, sizeof(manuf_name));
				write_result_string(resfd, 0, ret, manuf_name);
				write_result_string(resfd, 1, gpo_supported, model_name);  /* [sawas-5] */
				if(action != ACT_FILE_OPER_QUIT){
					action = ACT_NONE;
				}
			}

			switch (action) {
			case ACT_GET_DEVICE_INFO:
				write_result_string(resfd, 0, ret, manuf_name);
				write_result_string(resfd, 1, gpo_supported, model_name);  /* [sawas-5] */
				break;

			case ACT_GET_FILE:
				get_file(portn, busn,devn,force,handle,filename,overwrite,resfd);
				break;

#if  1  /* difference from original  [sawas-4] */
			case ACT_GET_PARTIAL_FILE:  /* for Part-OBJ  [sawas-55] */
#ifdef  DEBUG_PARTIAL
				printf("GET_PARTIAL_FILE: {port:0x%02x bus:%d dev:%d force:%d handle:%x}\n",  /* [sawas-6] */
											portn, busn, devn, force, handle);
				printf("GET_PARTIAL_FILE: {offset:0x%x size:0x%x overwrite:%d resfd:0x%x}\n",  /* [sawas-6] */
											offset, size, overwrite, resfd);
#endif
				if (size > part_size) {
					printf("GET_PARTIAL_FILE: size too big.\n");
					break;
				}
				get_partial_file(portn,busn,devn,force,handle,filename,offset,size,overwrite,resfd);
				break;
#endif
			case ACT_NUM_OF_OBJECTS:
				num_of_objects(portn, busn,devn,force,ofc,resfd);
				break;
			case ACT_LIST_EXIF_FILES:
				list_exif_files(portn, busn,devn,force,start,num,resfd);
				break;
			case ACT_GET_OBJECT_INFO:
				get_object_info (portn, busn,devn,force,handle,resfd);
				break;
			case ACT_GET_THUMB_FILE:
				get_thumb_file(portn, busn,devn,force,handle,filename,overwrite,resfd);
				break;
			case ACT_FILE_OPERATION:
				break;
			case ACT_FILE_OPER_QUIT:
				b_exit = 1;
				break;
			}

			if(action != ACT_NONE){
				action = ACT_NONE;
			}

		RES_END:
			// Output to file.
			if(resfd >= 0) {
				close(resfd);
			}
			unlink(FILE_OPERATION_RESLOCK);
		CMD_END:
			unlink(FILE_OPERATION_CMDLOCK);
		} // access(CMDFILE).

		if(b_exit){
			break;
		}
	} // end_while

	return;
}
#endif

/* main program  */

int
main(int argc, char ** argv)
{
	int portn=1,busn=0,devn=0;  /* [sawas-55] */
	int action=0;
	short force=0;
	int overwrite=SKIP_IF_EXISTS;
	uint16_t property=0;
	char* value=NULL;
	uint32_t handle=0;
#if  1  /* difference from original  [sawas-4] */
	char *execdir=NULL;
	uint32_t ofc=(uint32_t)-1;
	int resfd=-1;
	uint32_t offset=0;  /* for Part-OBJ  [sawas-55] */
#endif
	char *filename=NULL;
	int num=0;
	/* parse options */
	int option_index = 0,opt;
	static struct option loptions[] = {
		{"help",0,0,'h'},
		{"bus",1,0,0},
		{"dev",1,0,0},
#if  1  /* difference from original  [sawas-55] */
		{"hub",1,0,0},
		{"scan",1,0,0},
		{"port",1,0,0},
		{"size",1,0,0},
#endif
		{"reset",0,0,'r'},
		{"list-devices",0,0,'l'},
		{"list-files",0,0,'L'},
		{"list-operations",1,0,'o'},
		{"list-properties",0,0,'p'},
		{"show-property",1,0,'s'},
		{"set-property",1,0,'s'},
		{"get-file",1,0,'g'},
		{"get-all-files",0,0,'G'},
		{"capture",0,0,'c'},
		{"loop-capture",1,0,0},
		{"delete-object",1,0,'d'},
		{"delete-all-files",1,0,'D'},
		{"info",0,0,'i'},
		{"val",1,0,0},
		{"filename",1,0,0},
		{"overwrite",0,0,0},
		{"force",0,0,'f'},
		{"verbose",2,0,'v'},

#if  1  /* difference from original  [sawas-4] */
		{"num-objects",1,0,'n'},
		{"get-object-info",1,0,'I'},
		{"list-exif-files",0,0,'E'},
		{"get-thum-file",1,0,'t'},
		{"file-operation",1,0,'O'},
		{"get-partial-file",3,0,'P'},  /* for Part-OBJ  [sawas-55] */
#endif
		{0,0,0,0}
	};
	
#ifdef DEBUG  /* [sawas-6] */
	verbose = 1;
#endif

	while(1) {
		opt = getopt_long (argc, argv, "LhlcipfroGg:Dd:s:v::n:I:Et:O:P:", loptions, &option_index);  /* for Part-OBJ  [sawas-55] */
		if (opt==-1) break;
	
		switch (opt) {
		/* set parameters */
		case 0:
			if (!(strcmp("val",loptions[option_index].name)))
				value=strdup(optarg);
			if (!(strcmp("filename",loptions[option_index].name)))
				filename=strdup(optarg);
			if (!(strcmp("overwrite",loptions[option_index].name)))
				overwrite=OVERWRITE_EXISTING;
			if (!(strcmp("bus",loptions[option_index].name)))
				busn=strtol(optarg,NULL,10);
			if (!(strcmp("dev",loptions[option_index].name)))
				devn=strtol(optarg,NULL,10);
#if  1  /* difference from original  [sawas-55] */
			if (!(strcmp("hub",loptions[option_index].name))) {
				hub_id=strtol(optarg,NULL,16);
#ifdef  DEBUG_SEARCH
				printf("[ptpcam] hub_id=0x%08x\n", hub_id);
#endif
			}
			if (!(strcmp("scan",loptions[option_index].name))) {
#if  1  /* OHCI  [sawas-6] */
				scan_port=strtol(optarg,NULL,16);  /* [sawas-6] */
#ifdef  DEBUG_SEARCH
				printf("[ptpcam] scan  port(in )=0x%02x\n", scan_port );  /* [sawas-6] */
#endif
				/* get OHCI port */
				use_ohci2_port = scan_port & 0x20;  /* bit5 */
				scan_port &= 0xF;				/* lower 4bits */
#ifdef  DEBUG_SEARCH
				printf("[ptpcam] scan  port(out)=0x%02x\n", scan_port );  /* [sawas-6] */
				printf("[ptpcam] ohci2 port(use)=0x%02x\n", use_ohci2_port );  /* OHCI  [sawas-6] */
#endif
				if (use_ohci2_port) {
					printf("[ptpcam] use ohci2 port=0x%02x\n", use_ohci2_port);  /* OHCI  [sawas-6] */
				}
#endif
			}
			if (!(strcmp("port",loptions[option_index].name))) {
				portn=strtol(optarg,NULL,16);
#ifdef  DEBUG_SEARCH
				printf("[ptpcam] port number=0x%02x\n", portn);  /* sawas-6] */
#endif
				/* get hub port */
				hub_port = portn & 0x80;  /* bit 7  [sawas-6] */
				if (hub_port) {
					printf("[ptpcam] hub port number=0x%x\n", hub_port);
				}

				/* get OHCI#2 port  [sawas-6] */
				hub_port = portn & 0x20;  /* bit 5  [sawas-6] */
				if (hub_port) {
					printf("[ptpcam] hub port number=0x%x\n", hub_port);
				}

				/* get port */
				portn &= 0xF;  /* lower 4bits  [sawas-6] */
			}
			if (!(strcmp("size",loptions[option_index].name))) {
				part_size=strtol(optarg,NULL,16);
#ifdef  DEBUG_SEARCH
				printf("[ptpcam] partial size=0x%x\n", part_size);
#endif
			}
#endif
			if (!(strcmp("loop-capture",loptions[option_index].name)))
			{
				action=ACT_LOOP_CAPTURE;
				num=strtol(optarg,NULL,10);
			}

			break;
		case 'f':
			force=~force;
			break;
		case 'v':
			if (optarg) 
				verbose=strtol(optarg,NULL,10);
			else
				verbose=1;
			printf("VERBOSE LEVEL  = %i\n",verbose);
			break;

		/* actions */
		case 'h':
			help();
			break;
		case 'r':
			action=ACT_DEVICE_RESET;
			break;
		case 'l':
			action=ACT_LIST_DEVICES;
			break;
		case 'p':
			action=ACT_LIST_PROPERTIES;
			break;
		case 's':
			action=ACT_GETSET_PROPERTY;
			property=strtol(optarg,NULL,16);
			break;
		case 'o':
			action=ACT_LIST_OPERATIONS;
			break;
		case 'i':
			action=ACT_SHOW_INFO;
			break;
		case 'c':
			action=ACT_CAPTURE;
			break;
		case 'L':
			action=ACT_LIST_FILES;
			break;
		case 'g':
			action=ACT_GET_FILE;
			handle=strtol(optarg,NULL,16);
			break;
#if  1  /* for Part-OBJ  [sawas-55] */
		case 'P':
			action=ACT_GET_PARTIAL_FILE;
			handle=strtol(optarg,NULL,16);
			break;
#endif
		case 'G':
			action=ACT_GET_ALL_FILES;
			break;
		case 'd':
			action=ACT_DELETE_OBJECT;
			handle=strtol(optarg,NULL,16);
			break;
		case 'D':
			action=ACT_DELETE_ALL_FILES;

#if  1  /* difference from original  [sawas-4] */
			break;
		case 'n':
			action=ACT_NUM_OF_OBJECTS;
			ofc=strtol(optarg,NULL,16);
			break;
		case 'I':
			action=ACT_GET_OBJECT_INFO;
			handle=strtol(optarg,NULL,16);
			break;
		case 'E':
			action=ACT_LIST_EXIF_FILES;
			break;
		case 't':
			action=ACT_GET_THUMB_FILE;
			handle=strtol(optarg,NULL,16);
			break;
		case 'O':
			action=ACT_FILE_OPERATION;
			execdir=strdup(optarg);
			break;
#endif
		case '?':
			break;
		default:
			fprintf(stderr,"getopt returned character code 0%o\n",
				opt);
			break;
		}
	}

	if (argc==1) {
		usage();
		return 0;
	}

#if  0  /* difference from original  [sawas-5] */
	switch (action) {
		case ACT_DEVICE_RESET:
			reset_device(busn,devn,force);
			break;
		case ACT_LIST_DEVICES:
			list_devices(force);
			break;
		case ACT_LIST_PROPERTIES:
			list_properties(busn,devn,force);
			break;
		case ACT_GETSET_PROPERTY:
			getset_property(busn,devn,property,value,force);
			break;
		case ACT_SHOW_INFO:
			show_info(busn,devn,force);
			break;
		case ACT_LIST_OPERATIONS:
			list_operations(busn,devn,force);
			break;
		case ACT_LIST_FILES:
			list_files(busn,devn,force);
			break;
		case ACT_GET_FILE:
			get_file(busn,devn,force,handle,filename,overwrite);
			break;
		case ACT_GET_ALL_FILES:
			get_all_files(busn,devn,force,overwrite);
			break;
		case ACT_CAPTURE:
			capture_image(busn,devn,force);
			break;
		case ACT_DELETE_OBJECT:
			delete_object(busn,devn,force,handle);
			break;
		case ACT_DELETE_ALL_FILES:
			delete_all_files(busn,devn,force);
			break;
		case ACT_LOOP_CAPTURE:
			loop_capture (busn,devn,force,num,overwrite);
#else
	// set pid_file.
	const char *PID_FILENAME = "/var/tmp/ptpcam.pid";
	pid_t current_pid = getpid();
	FILE *fp = NULL;
	uint32_t readnum = 0;

	unlink(PID_FILENAME);
	if((fp = fopen( PID_FILENAME, "w" )) != NULL){
		fprintf(fp, "%d", current_pid);
		fclose(fp);
	}

	switch (action) {
		case ACT_DEVICE_RESET:
			reset_device(portn,busn,devn,force);
			break;
		case ACT_LIST_DEVICES:
			list_devices(force);
			break;
		case ACT_LIST_PROPERTIES:
			list_properties(portn,busn,devn,force);
			break;
		case ACT_GETSET_PROPERTY:
			getset_property(portn,busn,devn,property,value,force);
			break;
		case ACT_SHOW_INFO:
			show_info(portn,busn,devn,force,resfd);
			break;
		case ACT_LIST_OPERATIONS:
			list_operations(portn, busn,devn,force);
			break;
		case ACT_LIST_FILES:
			list_files(portn,busn,devn,force);
			break;
		case ACT_GET_FILE:
			get_file(portn,busn,devn,force,handle,filename,overwrite,resfd);
			break;

#if  1  /* for Part-OBJ  [sawas-55] */
		case ACT_GET_PARTIAL_FILE:
#ifdef  DEBUG_PARTIAL
			printf("GET_PARTIAL_FILE: {port:0x%02x bus:%d dev:%d force:%d handle:0x%x}\n",  /* [sawas-6] */
									portn, busn, devn, force, handle);  // [sawas-55]
			printf("GET_PART_FILE: {offset:0x%x size:0x%x overwrite:%d resfd:0x%x}\n",  /* [sawas-6] */
									offset, part_size, overwrite, resfd);  // [sawas-55]
			readnum = get_partial_file(portn,busn,devn,force,handle,filename,offset,part_size,overwrite,resfd);
			printf("GET_PART_FILE: {readnum:0x%x}\n",readnum);
#else
			readnum = get_partial_file(portn,busn,devn,force,handle,filename,offset,part_size,overwrite,resfd);
#endif
			break;
#endif
		case ACT_GET_ALL_FILES:
			get_all_files(portn, busn,devn,force,overwrite);
			break;
		case ACT_CAPTURE:
			capture_image(portn,busn,devn,force);
			break;
		case ACT_DELETE_OBJECT:
			delete_object(portn,busn,devn,force,handle);
			break;
		case ACT_DELETE_ALL_FILES:
			delete_all_files(portn, busn,devn,force);
			break;
		case ACT_LOOP_CAPTURE:
			loop_capture (portn, busn,devn,force,num,overwrite);
			break;
		case ACT_NUM_OF_OBJECTS:
			num_of_objects(portn,busn,devn,force, ofc, resfd);
			break;
		case ACT_LIST_EXIF_FILES:
			list_exif_files (portn,busn,devn,force,0,0,resfd);
			break;
		case ACT_GET_OBJECT_INFO:
			get_object_info (portn,busn,devn,force, handle, resfd);
			break;
		case ACT_GET_THUMB_FILE:
			get_thumb_file(portn,busn,devn,force,handle,filename,overwrite,resfd);
			break;
		case ACT_FILE_OPERATION:
			file_operation (portn,busn,devn,force, execdir);
			break;
#endif
	}

	return 0;
}
