/* ptpcam.h
 *
 * Copyright (C) 2001-2005 Mariusz Woloszyn <emsi@ipartners.pl>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 */
#ifndef __PTPCAM_H__
#define __PTPCAM_H__

#ifdef LINUX_OS
#define USB_BULK_READ myusb_bulk_read
#define USB_BULK_WRITE myusb_bulk_write
int myusb_bulk_read(usb_dev_handle *dev, int ep, char *bytes, int size,
	int timeout);
int myusb_bulk_write(usb_dev_handle *dev, int ep, char *bytes, int length,
	int timeout);
#else
#define USB_BULK_READ usb_bulk_read
#define USB_BULK_WRITE usb_bulk_write
#endif

/*
 * macros
 */

/* Check value and Return on error */
#define CR(result,error) {						\
			if((result)!=PTP_RC_OK) {			\
				fprintf(stderr,"ERROR: "error);		\
				close_camera(&ptp_usb, &params, dev);   \
				return;					\
			}						\
}

/* Check value and Continue on error */
#define CC(result,error) {						\
			if((result)!=PTP_RC_OK) {			\
				fprintf(stderr,"ERROR: "error);		\
				usb_release_interface(ptp_usb.handle,	\
		dev->config->interface->altsetting->bInterfaceNumber);	\
				continue;					\
			}						\
}

/* error reporting macro */
#define ERROR(error) fprintf(stderr,"ERROR: "error);				

/* property value printing macros */
#define PRINT_PROPVAL_DEC(value)	\
		print_propval(dpd.DataType, value,			\
		PTPCAM_PRINT_DEC)

#define PRINT_PROPVAL_HEX(value)					\
		print_propval(dpd.DataType, value,			\
		PTPCAM_PRINT_HEX)




/*
 * defines
 */
#define  BUF_SIZE  140  /* [sawas-55] */

/* requested actions */
#define ACT_DEVICE_RESET		0x1
#define ACT_LIST_DEVICES		0x2
#define ACT_LIST_PROPERTIES		0x3
#define ACT_LIST_OPERATIONS		0x4
#define ACT_GETSET_PROPERTY		0x5
#define ACT_SHOW_INFO			0x6
#define ACT_LIST_FILES			0x7
#define ACT_GET_FILE			0x8
#define ACT_GET_ALL_FILES		0x9
#define ACT_CAPTURE				0xA
#define ACT_DELETE_OBJECT		0xB
#define ACT_DELETE_ALL_FILES	0xC
#define ACT_LOOP_CAPTURE		0xD

#if  1  /* difference from original  [sawas-4] */
#define ACT_GET_DEVICE_INFO		0x80
#define ACT_NUM_OF_OBJECTS		0x81
#define ACT_LIST_EXIF_FILES		0x82
#define ACT_GET_OBJECT_INFO		0x83
#define ACT_GET_THUMB_FILE		0x84
#define ACT_FILE_OPERATION		0xF1
#define ACT_GET_PARTIAL_FILE	0xF2  /* [sawas-55] */
#define ACT_FILE_OPER_QUIT		0xF9
#define ACT_NONE				0x00

#define FILE_OPERATION_CMDFILE "ptp.cmd"
#define FILE_OPERATION_CMDLOCK "ptp.cmd.lck"
#define FILE_OPERATION_RESFILE "ptp.res"
#define FILE_OPERATION_RESLOCK "ptp.res.lck"
#define FILE_OPERATION_CMDLAST "ptp.cmd.last"
/* default get_file filename (HANDLE_FILENAME) */
#define GET_FILE_DEFAULT_FMT   "%08x_%s"
#define GET_THUMB_DEFAULT_FMT   "t%08x_%s"
#endif
/* printing value type */
#define PTPCAM_PRINT_HEX	00
#define PTPCAM_PRINT_DEC	01

/* filename overwrite */
#define OVERWRITE_EXISTING	1
#define	SKIP_IF_EXISTS		0

/* number of port */
#define MAX_PORT_NUMBER		3  /* [sawas-5] */


/*
 * structures
 */

typedef struct _PTP_USB PTP_USB;
struct _PTP_USB {
	usb_dev_handle* handle;
	int inep;
	int outep;
	int intep;
};

/*
 * variables
 */

/* one global variable */
extern short verbose;


/*
 * functions
 */

void usage(void);
void help(void);

int GetDeviceDescType(char *lineBuf);  /* [sawas-5] */
int SearchDevice(int portno, int busno, int devno);  /* [sawas-5] */

void list_devices(short force);
void show_info (int portn, int busn, int devn, short force, int resfd);  /* [sawas-5] */
void list_files (int portn, int busn, int devn, short force);  /* [sawas-5] */
void get_file (int portn, int busn, int devn, short force, uint32_t handle, char* filename,
			   int overwrite, int resfd);  /* [sawas-5] */
uint32_t get_partial_file (int portn, int busn, int devn, short force, uint32_t handle, char* filename,
					   uint32_t offset, uint32_t size, int overwrite, int resfd);  /* [sawas-55] */
void get_all_files (int portn, int busn, int devn, short force, int overwrite);  /* [sawas-5] */
void capture_image (int portn, int busn, int devn, short force);  /* [sawas-5] */
void delete_object (int portn, int busn, int devn, short force, uint32_t handle);  /* [sawas-5] */
void delete_all_files (int portn, int busn, int devn, short force);  /* [sawas-5] */
void list_operations (int portn, int busn, int devn, short force);  /* [sawas-5] */
void list_devices(short force);
void list_properties (int portn, int busn, int devn, short force);  /* [sawas-5] */
void loop_capture (int portn, int busn, int devn, short force, int n,  int overwrite);  /* [sawas-5] */

struct usb_bus* init_usb(void);
void close_usb(PTP_USB* ptp_usb, struct usb_device* dev);
void init_ptp_usb (PTPParams*, PTP_USB*, struct usb_device*);
void clear_stall(PTP_USB* ptp_usb, struct usb_device* dev);

int usb_get_endpoint_status(PTP_USB* ptp_usb, int ep, uint16_t* status);
int usb_clear_stall_feature(PTP_USB* ptp_usb, int ep);

int open_camera (int portn, int busn, int devn, short force, PTP_USB *ptp_usb, PTPParams *params, struct usb_device **dev);  /* [sawas-5] */
void close_camera (PTP_USB *ptp_usb, PTPParams *params, struct usb_device *dev);
int is_gpo_supported(PTPParams *params);  /* for Part-OBJ  [sawas-55] */
int get_info (int portn, int busn, int devn, short force, char *model, int model_size, char *manuf, int manuf_size, int *gpo_supported);  /* [sawas-55] */
void print_object_info(int n, uint32_t handle, const PTPObjectInfo *oi, int vflag);
void num_of_objects (int portn, int busn, int devn, short force, uint32_t ofc, int resfd);  /* [sawas-5] */
void list_exif_files (int portn, int busn, int devn, short force, uint32_t start, uint32_t num, int resfd);  /* [sawas-5] */
void get_object_info (int portn, int busn, int devn, short force, uint32_t handle, int resfd);  /* [sawas-5] */
void get_thumb_file (int portn, int busn, int devn, short force, uint32_t handle, char* filename, int overwrite, int resfd);  /* [sawas-5] */
void file_operation (int portn, int busn, int devn, short force, char* execdir);  /* [sawas-5] */
void write_result_head(int fd, int action, uint32_t trid, uint32_t param1, uint32_t param2, uint32_t param3, const char *model_name);  /* for Part-OBJ  [sawas-55] */
void write_result(int fd, int num, int errcode, uint32_t handle, const PTPObjectInfo *oi);
void write_result_string(int fd, int num, int errcode, const char *res_str);

#endif /* __PTPCAM_H__ */
